"""
834.4010.X095.A1
Created 2023-05-19 10:17:35.956938
"""
from .base import *
from .common import *


class ISA_LOOP_ISA(Segment):
    """Interchange Control Header"""
    _segment_name = 'ISA'
    
    isa01: Annotated[I01, Title('Authorization Information Qualifier'), Usage('R'), Position(1), Enumerated(*['00', '03'])]
    
    isa02: Annotated[I02, Title('Authorization Information'), Usage('R'), Position(2)]
    
    isa03: Annotated[I03, Title('Security Information Qualifier'), Usage('R'), Position(3), Enumerated(*['00', '01'])]
    
    isa04: Annotated[I04, Title('Security Information'), Usage('R'), Position(4)]
    
    isa05: Annotated[I05, Title('Interchange ID Qualifier'), Usage('R'), Position(5), Enumerated(*['01', '14', '20', '27', '28', '29', '30', '33', 'ZZ'])]
    
    isa06: Annotated[I06, Title('Interchange Sender ID'), Usage('R'), Position(6)]
    
    isa07: Annotated[I05, Title('Interchange ID Qualifier'), Usage('R'), Position(7), Enumerated(*['01', '14', '20', '27', '28', '29', '30', '33', 'ZZ'])]
    
    isa08: Annotated[I07, Title('Interchange Receiver ID'), Usage('R'), Position(8)]
    
    isa09: Annotated[I08, Title('Interchange Date'), Usage('R'), Position(9)]
    
    isa10: Annotated[I09, Title('Interchange Time'), Usage('R'), Position(10)]
    
    isa11: Annotated[I10, Title('Interchange Control Standards Identifier'), Usage('R'), Position(11), Enumerated(*['U'])]
    
    isa12: Annotated[I11, Title('Interchange Control Version Number'), Usage('R'), Position(12), Enumerated(*['00401'])]
    
    isa13: Annotated[I12, Title('Interchange Control Number'), Usage('R'), Position(13)]
    
    isa14: Annotated[I13, Title('Acknowledgment Requested'), Usage('R'), Position(14), Enumerated(*['0', '1'])]
    
    isa15: Annotated[I14, Title('Usage Indicator'), Usage('R'), Position(15), Enumerated(*['P', 'T'])]
    
    isa16: Annotated[I15, Title('Component Element Separator'), Usage('R'), Position(16)]


class GS_LOOP_GS(Segment):
    """Functional Group Header"""
    _segment_name = 'GS'
    
    gs01: Annotated[D_479, Title('Functional Identifier Code'), Usage('R'), Position(1), Enumerated(*['BE'])]
    
    gs02: Annotated[D_142, Title("Application Sender's Code"), Usage('R'), Position(2)]
    
    gs03: Annotated[D_124, Title("Application Receiver's Code"), Usage('R'), Position(3)]
    
    gs04: Annotated[D_373, Title('Date'), Usage('R'), Position(4)]
    
    gs05: Annotated[D_337, Title('Time'), Usage('R'), Position(5)]
    
    gs06: Annotated[D_28, Title('Group Control Number'), Usage('R'), Position(6)]
    
    gs07: Annotated[D_455, Title('Responsible Agency Code'), Usage('R'), Position(7), Enumerated(*['X'])]
    
    gs08: Annotated[D_480, Title('Version / Release / Industry Identifier Code'), Usage('R'), Position(8), Enumerated(*['004010X095A1'])]


class ST_LOOP_ST(Segment):
    """Transaction Set Header"""
    _segment_name = 'ST'
    
    st01: Annotated[D_143, Title('Transaction Set Identifier Code'), Usage('R'), Position(1), Enumerated(*['834'])]
    
    st02: Annotated[D_329, Title('Transaction Set Control Number'), Usage('R'), Position(2)]


class HEADER_BGN(Segment):
    """Beginning Segment"""
    _segment_name = 'BGN'
    
    bgn01: Annotated[D_353, Title('Transaction Set Purpose Code'), Usage('R'), Position(1), Enumerated(*['00', '15', '22'])]
    
    bgn02: Annotated[D_127, Title('Transaction Set Identifier Code'), Usage('R'), Position(2)]
    
    bgn03: Annotated[D_373, Title('Transaction Set Creation Date'), Usage('R'), Position(3)]
    
    bgn04: Annotated[D_337, Title('Transaction Set Creation Time'), Usage('R'), Position(4)]
    
    bgn05: Annotated[D_623, Title('Time Zone Code'), Usage('S'), Position(5), Enumerated(*['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', 'AD', 'AS', 'AT', 'CD', 'CS', 'CT', 'ED', 'ES', 'ET', 'GM', 'HD', 'HS', 'HT', 'LT', 'MD', 'MS', 'MT', 'ND', 'NS', 'NT', 'PD', 'PS', 'PT', 'TD', 'TS', 'TT', 'UT'])]
    
    bgn06: Annotated[D_127, Title('Transaction Set Identifier Code'), Usage('S'), Position(6)]
    
    bgn07: Annotated[D_640, Title('Transaction Type Code'), Usage('N'), Position(7)]
    
    bgn08: Annotated[D_306, Title('Action Code'), Usage('R'), Position(8), Enumerated(*['2', '4'])]
    
    bgn09: Annotated[D_786, Title('Security Level Code'), Usage('N'), Position(9)]


class HEADER_C040(Composite):
    """Reference Identifier"""
    pass


class HEADER_REF(Segment):
    """Transaction Set Policy Number"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['38'])]
    
    ref02: Annotated[D_127, Title('Master Policy Number'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class HEADER_DTP(Segment):
    """File Effective Date"""
    _segment_name = 'DTP'
    
    dtp01: Annotated[D_374, Title('Date Time Qualifier'), Usage('R'), Position(1), Enumerated(*['007', '303', '382', '388'])]
    
    dtp02: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(2), Enumerated(*['D8'])]
    
    dtp03: Annotated[D_1251, Title('Date Time Period'), Usage('R'), Position(3)]


class L1000A_N1(Segment):
    """Sponsor Name"""
    _segment_name = 'N1'
    
    n101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['P5'])]
    
    n102: Annotated[D_93, Title('Plan Sponsor Name'), Usage('S'), Position(2)]
    
    n103: Annotated[D_66, Title('Identification Code Qualifier'), Usage('R'), Position(3), Enumerated(*['FI', 'ZZ'])]
    
    n104: Annotated[D_67, Title('Sponsor Identifier'), Usage('R'), Position(4)]
    
    n105: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(5)]
    
    n106: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(6)]


class L1000A(Loop):
    
    n1: Annotated[L1000A_N1, Title('Sponsor Name'), Usage('R'), Position(70), Syntax(['R0203', 'P0304']), Required(True)]


class L1000B_N1(Segment):
    """Payer"""
    _segment_name = 'N1'
    
    n101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['IN'])]
    
    n102: Annotated[D_93, Title('Insurer Name'), Usage('S'), Position(2)]
    
    n103: Annotated[D_66, Title('Identification Code Qualifier'), Usage('R'), Position(3), Enumerated(*['FI', 'XV'])]
    
    n104: Annotated[D_67, Title('Insurer Identification Code'), Usage('R'), Position(4)]
    
    n105: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(5)]
    
    n106: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(6)]


class L1000B(Loop):
    
    n1: Annotated[L1000B_N1, Title('Payer'), Usage('R'), Position(70), Syntax(['R0203', 'P0304']), Required(True)]


class L1000C_N1(Segment):
    """TPA/Broker Name"""
    _segment_name = 'N1'
    
    n101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['BO', 'TV'])]
    
    n102: Annotated[D_93, Title('TPA or Broker Name'), Usage('R'), Position(2)]
    
    n103: Annotated[D_66, Title('Identification Code Qualifier'), Usage('R'), Position(3), Enumerated(*['94', 'FI', 'XV'])]
    
    n104: Annotated[D_67, Title('TPA or Broker Identification Code'), Usage('R'), Position(4)]
    
    n105: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(5)]
    
    n106: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(6)]


class L1100C_ACT(Segment):
    """TPA/Broker Account Information"""
    _segment_name = 'ACT'
    
    act01: Annotated[D_508, Title('TPA or Broker Account Number'), Usage('R'), Position(1)]
    
    act02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    act03: Annotated[D_66, Title('Identification Code Qualifier'), Usage('N'), Position(3)]
    
    act04: Annotated[D_67, Title('Identification Code'), Usage('N'), Position(4)]
    
    act05: Annotated[D_569, Title('Account Number Qualifier'), Usage('N'), Position(5)]
    
    act06: Annotated[D_508, Title('TPA or Broker Account Number'), Usage('S'), Position(6)]
    
    act07: Annotated[D_352, Title('Description'), Usage('N'), Position(7)]
    
    act08: Annotated[D_107, Title('Payment Method Code'), Usage('N'), Position(8)]
    
    act09: Annotated[D_1216, Title('Benefit Status Code'), Usage('N'), Position(9)]


class L1100C(Loop):
    
    act: Annotated[L1100C_ACT, Title('TPA/Broker Account Information'), Usage('R'), Position(120), Syntax(['P0304', 'C0506', 'C0705']), Required(True)]


class L1000C(Loop):
    
    n1: Annotated[L1000C_N1, Title('TPA/Broker Name'), Usage('R'), Position(70), Syntax(['R0203', 'P0304']), Required(True)]
    ItemL1100C: TypeAlias = Annotated[L1100C, Title('TPA/Broker Account Information'), Usage('S'), Position(120), Required(True)]
    l1100c: Annotated[list[ItemL1100C], MinItems(1)]


class HEADER(Loop):
    
    bgn: Annotated[HEADER_BGN, Title('Beginning Segment'), Usage('R'), Position(20), Syntax(['C0504']), Required(True)]
    
    ref: Annotated[HEADER_REF, Title('Transaction Set Policy Number'), Usage('S'), Position(30), Syntax(['R0203']), Required(True)]
    ItemDtp: TypeAlias = Annotated[HEADER_DTP, Title('File Effective Date'), Usage('S'), Position(40), Required(True)]
    dtp: Annotated[list[ItemDtp], MinItems(1)]
    ItemL1000A: TypeAlias = Annotated[L1000A, Title('Sponsor Name'), Usage('R'), Position(70), Required(True)]
    l1000a: Annotated[list[ItemL1000A], MinItems(1)]
    ItemL1000B: TypeAlias = Annotated[L1000B, Title('Payer'), Usage('R'), Position(70), Required(True)]
    l1000b: Annotated[list[ItemL1000B], MinItems(1)]
    ItemL1000C: TypeAlias = Annotated[L1000C, Title('TPA/Broker Name'), Usage('S'), Position(70), Required(True)]
    l1000c: Annotated[list[ItemL1000C], MinItems(2)]


class L2000_INS(Segment):
    """Member Level Detail"""
    _segment_name = 'INS'
    
    ins01: Annotated[D_1073, Title('Insured Indicator'), Usage('R'), Position(1), Enumerated(*['N', 'Y'])]
    
    ins02: Annotated[D_1069, Title('Individual Relationship Code'), Usage('R'), Position(2), Enumerated(*['01', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12', '13', '14', '15', '17', '18', '19', '23', '24', '25', '26', '31', '32', '33', '38', '48', '49', '53'])]
    
    ins03: Annotated[D_875, Title('Maintenance Type Code'), Usage('R'), Position(3), Enumerated(*['001', '021', '024', '025', '030'])]
    
    ins04: Annotated[D_1203, Title('Maintenance Reason Code'), Usage('S'), Position(4), Enumerated(*['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '14', '15', '16', '17', '18', '20', '21', '22', '25', '26', '27', '28', '29', '31', '32', '33', '37', '38', '39', '40', '41', '43', 'AI', 'XN', 'XT'])]
    
    ins05: Annotated[D_1216, Title('Benefit Status Code'), Usage('R'), Position(5), Enumerated(*['A', 'C', 'S', 'T'])]
    
    ins06: Annotated[D_1218, Title('Medicare Plan Code'), Usage('S'), Position(6), Enumerated(*['A', 'B', 'C', 'D', 'E'])]
    
    ins07: Annotated[D_1219, Title('Consolidated Omnibus Budget Reconciliation Act (COBRA) Qualifying Event Code'), Usage('S'), Position(7), Enumerated(*['1', '2', '3', '4', '5', '6', '7', '8'])]
    
    ins08: Annotated[D_584, Title('Employment Status Code'), Usage('S'), Position(8), Enumerated(*['AO', 'AU', 'FT', 'L1', 'PT', 'RT', 'TE'])]
    
    ins09: Annotated[D_1220, Title('Student Status Code'), Usage('S'), Position(9), Enumerated(*['F', 'N', 'P'])]
    
    ins10: Annotated[D_1073, Title('Handicap Indicator'), Usage('S'), Position(10), Enumerated(*['N', 'Y'])]
    
    ins11: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('S'), Position(11), Enumerated(*['D8'])]
    
    ins12: Annotated[D_1251, Title('Insured Individual Death Date'), Usage('S'), Position(12)]
    
    ins13: Annotated[D_1165, Title('Confidentiality Code'), Usage('N'), Position(13)]
    
    ins14: Annotated[D_19, Title('City Name'), Usage('N'), Position(14)]
    
    ins15: Annotated[D_156, Title('State or Province Code'), Usage('N'), Position(15)]
    
    ins16: Annotated[D_26, Title('Country Code'), Usage('N'), Position(16)]
    
    ins17: Annotated[D_1470, Title('Birth Sequence Number'), Usage('S'), Position(17)]


class L2000_C040(Composite):
    """Reference Identifier"""
    pass


class L2000_REF(Segment):
    """Subscriber Number"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['0F'])]
    
    ref02: Annotated[D_127, Title('Subscriber Identifier'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class L2000_C040(Composite):
    """Reference Identifier"""
    pass


class L2000_REF(Segment):
    """Member Policy Number"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['1L'])]
    
    ref02: Annotated[D_127, Title('Insured Group or Policy Number'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class L2000_C040(Composite):
    """Reference Identifier"""
    pass


class L2000_REF(Segment):
    """Member Identification Number"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['17', '23', '3H', '6O', 'DX', 'F6', 'Q4', 'ZZ'])]
    
    ref02: Annotated[D_127, Title('Subscriber Supplemental Identifier'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class L2000_C040(Composite):
    """Reference Identifier"""
    pass


class L2000_REF(Segment):
    """Prior Coverage Months"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['QQ'])]
    
    ref02: Annotated[D_127, Title('Prior Coverage Month Count'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class L2000_DTP(Segment):
    """Member Level Dates"""
    _segment_name = 'DTP'
    
    dtp01: Annotated[D_374, Title('Date Time Qualifier'), Usage('R'), Position(1), Enumerated(*['286', '296', '297', '300', '301', '303', '336', '337', '338', '339', '340', '341', '350', '351', '356', '357', '383', '393', '394', '473', '474'])]
    
    dtp02: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(2), Enumerated(*['D8'])]
    
    dtp03: Annotated[D_1251, Title('Status Information Effective Date'), Usage('R'), Position(3)]


class L2100A_NM1(Segment):
    """Member Name"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['74', 'IL'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1'])]
    
    nm103: Annotated[D_1035, Title('Subscriber Last Name'), Usage('R'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Subscriber First Name'), Usage('R'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Subscriber Middle Name'), Usage('S'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Subscriber Name Prefix'), Usage('S'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Subscriber Name Suffix'), Usage('S'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(8), Enumerated(*['34', 'ZZ'])]
    
    nm109: Annotated[D_67, Title('Subscriber Identifier'), Usage('S'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100A_PER(Segment):
    """Member Communications Numbers"""
    _segment_name = 'PER'
    
    per01: Annotated[D_366, Title('Contact Function Code'), Usage('R'), Position(1), Enumerated(*['IP'])]
    
    per02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    per03: Annotated[D_365, Title('Communication Number Qualifier'), Usage('R'), Position(3), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per04: Annotated[D_364, Title('Communication Number'), Usage('R'), Position(4)]
    
    per05: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(5), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per06: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(6)]
    
    per07: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(7), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per08: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(8)]
    
    per09: Annotated[D_443, Title('Contact Inquiry Reference'), Usage('N'), Position(9)]


class L2100A_N3(Segment):
    """Member Residence Street Address"""
    _segment_name = 'N3'
    
    n301: Annotated[D_166, Title('Subscriber Address Line'), Usage('R'), Position(1)]
    
    n302: Annotated[D_166, Title('Subscriber Address Line'), Usage('S'), Position(2)]


class L2100A_N4(Segment):
    """Member Residence City, State, Zip Code"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('Subscriber City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('Subscriber State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('Subscriber Postal Zone or ZIP Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('S'), Position(5), Enumerated(*['60', 'CY'])]
    
    n406: Annotated[D_310, Title('Location Identification Code'), Usage('S'), Position(6)]


class L2100A_DMG(Segment):
    """Member Demographics"""
    _segment_name = 'DMG'
    
    dmg01: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(1), Enumerated(*['D8'])]
    
    dmg02: Annotated[D_1251, Title('Member Birth Date'), Usage('R'), Position(2)]
    
    dmg03: Annotated[D_1068, Title('Gender Code'), Usage('R'), Position(3), Enumerated(*['F', 'M', 'U'])]
    
    dmg04: Annotated[D_1067, Title('Marital Status Code'), Usage('S'), Position(4), Enumerated(*['B', 'D', 'I', 'M', 'R', 'S', 'U', 'W', 'X'])]
    
    dmg05: Annotated[D_1109, Title('Race or Ethnicity Code'), Usage('S'), Position(5), Enumerated(*['7', '8', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'N', 'O', 'P', 'Z'])]
    
    dmg06: Annotated[D_1066, Title('Citizenship Status Code'), Usage('S'), Position(6), Enumerated(*['1', '2', '3', '4', '5', '6', '7'])]
    
    dmg07: Annotated[D_26, Title('Country Code'), Usage('N'), Position(7)]
    
    dmg08: Annotated[D_659, Title('Basis of Verification Code'), Usage('N'), Position(8)]
    
    dmg09: Annotated[D_380, Title('Quantity'), Usage('N'), Position(9)]


class L2100A_ICM(Segment):
    """Member Income"""
    _segment_name = 'ICM'
    
    icm01: Annotated[D_594, Title('Frequency Code'), Usage('R'), Position(1), Enumerated(*['1', '2', '3', '4', '6', '7', '8', '9', 'B', 'C', 'H', 'Q', 'S', 'U'])]
    
    icm02: Annotated[D_782, Title('Wage Amount'), Usage('R'), Position(2)]
    
    icm03: Annotated[D_380, Title('Work Hours Count'), Usage('S'), Position(3)]
    
    icm04: Annotated[D_310, Title('Location Identification Code'), Usage('S'), Position(4)]
    
    icm05: Annotated[D_1214, Title('Salary Grade Code'), Usage('S'), Position(5)]
    
    icm06: Annotated[D_100, Title('Currency Code'), Usage('N'), Position(6)]


class L2100A_AMT(Segment):
    """Member Policy Amounts"""
    _segment_name = 'AMT'
    
    amt01: Annotated[D_522, Title('Amount Qualifier Code'), Usage('R'), Position(1), Enumerated(*['B9', 'C1', 'D2', 'P3'])]
    
    amt02: Annotated[D_782, Title('Contract Amount'), Usage('R'), Position(2)]
    
    amt03: Annotated[D_478, Title('Credit/Debit Flag Code'), Usage('N'), Position(3)]


class L2100A_HLH(Segment):
    """Member Health Information"""
    _segment_name = 'HLH'
    
    hlh01: Annotated[D_1212, Title('Health Related Code'), Usage('S'), Position(1), Enumerated(*['N', 'S', 'T', 'U', 'X'])]
    
    hlh02: Annotated[D_65, Title('Member Height'), Usage('S'), Position(2)]
    
    hlh03: Annotated[D_81, Title('Member Weight'), Usage('S'), Position(3)]
    
    hlh04: Annotated[D_81, Title('Weight'), Usage('N'), Position(4)]
    
    hlh05: Annotated[D_352, Title('Description'), Usage('N'), Position(5)]
    
    hlh06: Annotated[D_1213, Title('Current Health Condition Code'), Usage('N'), Position(6)]
    
    hlh07: Annotated[D_352, Title('Description'), Usage('N'), Position(7)]


class L2100A_LUI(Segment):
    """Member Language"""
    _segment_name = 'LUI'
    
    lui01: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(1), Enumerated(*['LD', 'LE'])]
    
    lui02: Annotated[D_67, Title('Language Code'), Usage('S'), Position(2)]
    
    lui03: Annotated[D_352, Title('Language Description'), Usage('S'), Position(3)]
    
    lui04: Annotated[D_1303, Title('Language Use Indicator'), Usage('S'), Position(4), Enumerated(*['5', '7', '8'])]
    
    lui05: Annotated[D_1476, Title('Language Proficiency Indicator'), Usage('N'), Position(5)]


class L2100A(Loop):
    
    nm1: Annotated[L2100A_NM1, Title('Member Name'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    per: Annotated[L2100A_PER, Title('Member Communications Numbers'), Usage('S'), Position(40), Syntax(['P0304', 'P0506', 'P0708']), Required(True)]
    
    n3: Annotated[L2100A_N3, Title('Member Residence Street Address'), Usage('S'), Position(50), Required(True)]
    
    n4: Annotated[L2100A_N4, Title('Member Residence City, State, Zip Code'), Usage('S'), Position(60), Syntax(['C0605']), Required(True)]
    
    dmg: Annotated[L2100A_DMG, Title('Member Demographics'), Usage('S'), Position(80), Syntax(['P0102']), Required(True)]
    
    icm: Annotated[L2100A_ICM, Title('Member Income'), Usage('S'), Position(110), Required(True)]
    ItemAmt: TypeAlias = Annotated[L2100A_AMT, Title('Member Policy Amounts'), Usage('S'), Position(120), Required(True)]
    amt: Annotated[list[ItemAmt], MaxItems(4)]
    
    hlh: Annotated[L2100A_HLH, Title('Member Health Information'), Usage('S'), Position(130)]
    ItemLui: TypeAlias = Annotated[L2100A_LUI, Title('Member Language'), Usage('S'), Position(150), Syntax(['P0102', 'L040203'])]
    lui: Annotated[list[ItemLui], MaxItems(5)]


class L2100B_NM1(Segment):
    """Incorrect Member Name"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['70'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1'])]
    
    nm103: Annotated[D_1035, Title('Prior Incorrect Insured Last Name'), Usage('R'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Prior Incorrect Insured First Name'), Usage('R'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Prior Incorrect Insured Middle Name'), Usage('S'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Prior Incorrect Insured Name Prefix'), Usage('S'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Prior Incorrect Insured Name Suffix'), Usage('S'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(8), Enumerated(*['34', 'ZZ'])]
    
    nm109: Annotated[D_67, Title('Prior Incorrect Insured Identifier'), Usage('S'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100B_DMG(Segment):
    """Incorrect Member Demographics"""
    _segment_name = 'DMG'
    
    dmg01: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(1), Enumerated(*['D8'])]
    
    dmg02: Annotated[D_1251, Title('Prior Incorrect Insured Birth Date'), Usage('R'), Position(2)]
    
    dmg03: Annotated[D_1068, Title('Prior Incorrect Insured Gender Code'), Usage('R'), Position(3), Enumerated(*['F', 'M', 'U'])]
    
    dmg04: Annotated[D_1067, Title('Marital Status Code'), Usage('N'), Position(4)]
    
    dmg05: Annotated[D_1109, Title('Race or Ethnicity Code'), Usage('N'), Position(5)]
    
    dmg06: Annotated[D_1066, Title('Citizenship Status Code'), Usage('N'), Position(6)]
    
    dmg07: Annotated[D_26, Title('Country Code'), Usage('N'), Position(7)]
    
    dmg08: Annotated[D_659, Title('Basis of Verification Code'), Usage('N'), Position(8)]
    
    dmg09: Annotated[D_380, Title('Quantity'), Usage('N'), Position(9)]


class L2100B(Loop):
    
    nm1: Annotated[L2100B_NM1, Title('Incorrect Member Name'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    dmg: Annotated[L2100B_DMG, Title('Incorrect Member Demographics'), Usage('S'), Position(80), Syntax(['P0102']), Required(True)]


class L2100C_NM1(Segment):
    """Member Mailing Address"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['31'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1'])]
    
    nm103: Annotated[D_1035, Title('Name Last or Organization Name'), Usage('N'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Name First'), Usage('N'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Name Middle'), Usage('N'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Name Prefix'), Usage('N'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Name Suffix'), Usage('N'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('N'), Position(8)]
    
    nm109: Annotated[D_67, Title('Identification Code'), Usage('N'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100C_N3(Segment):
    """Member Mail Street Address"""
    _segment_name = 'N3'
    
    n301: Annotated[D_166, Title('Subscriber Address Line'), Usage('R'), Position(1)]
    
    n302: Annotated[D_166, Title('Subscriber Address Line'), Usage('S'), Position(2)]


class L2100C_N4(Segment):
    """Member Mail City, State, Zip"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('Subscriber City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('Subscriber State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('Subscriber Postal Zone or ZIP Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4), Enumerated(*country)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('N'), Position(5)]
    
    n406: Annotated[D_310, Title('Location Identifier'), Usage('N'), Position(6)]


class L2100C(Loop):
    
    nm1: Annotated[L2100C_NM1, Title('Member Mailing Address'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    n3: Annotated[L2100C_N3, Title('Member Mail Street Address'), Usage('S'), Position(50), Required(True)]
    
    n4: Annotated[L2100C_N4, Title('Member Mail City, State, Zip'), Usage('S'), Position(60), Syntax(['C0605']), Required(True)]


class L2100D_NM1(Segment):
    """Member Employer"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['ES'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1', '2'])]
    
    nm103: Annotated[D_1035, Title('Insured Employer Name'), Usage('S'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Insured Employer First Name'), Usage('S'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Insured Employer Middle Name'), Usage('S'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Name Prefix'), Usage('N'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Insured Employer Name Suffix'), Usage('S'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(8), Enumerated(*['ZZ'])]
    
    nm109: Annotated[D_67, Title('Insured Employer Identifier'), Usage('S'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100D_PER(Segment):
    """Member Employer Communications Numbers"""
    _segment_name = 'PER'
    
    per01: Annotated[D_366, Title('Contact Function Code'), Usage('R'), Position(1), Enumerated(*['EP'])]
    
    per02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    per03: Annotated[D_365, Title('Communication Number Qualifier'), Usage('R'), Position(3), Enumerated(*['EM', 'EX', 'FX', 'TE'])]
    
    per04: Annotated[D_364, Title('Communication Number'), Usage('R'), Position(4)]
    
    per05: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(5), Enumerated(*['EM', 'EX', 'FX', 'TE'])]
    
    per06: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(6)]
    
    per07: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(7), Enumerated(*['EM', 'EX', 'FX', 'TE'])]
    
    per08: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(8)]
    
    per09: Annotated[D_443, Title('Contact Inquiry Reference'), Usage('N'), Position(9)]


class L2100D_N3(Segment):
    """Member Employer Street Address"""
    _segment_name = 'N3'
    
    n301: Annotated[D_166, Title('Insured Employer Address Line'), Usage('R'), Position(1)]
    
    n302: Annotated[D_166, Title('Insured Employer Address Line'), Usage('S'), Position(2)]


class L2100D_N4(Segment):
    """Member Employer City, State, Zip"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('Insured Employer City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('Insured Employer State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('Insured Employer Postal Zone or ZIP Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4), Enumerated(*country)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('N'), Position(5)]
    
    n406: Annotated[D_310, Title('Location Identifier'), Usage('N'), Position(6)]


class L2100D(Loop):
    
    nm1: Annotated[L2100D_NM1, Title('Member Employer'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    per: Annotated[L2100D_PER, Title('Member Employer Communications Numbers'), Usage('S'), Position(40), Syntax(['P0304', 'P0506', 'P0708']), Required(True)]
    
    n3: Annotated[L2100D_N3, Title('Member Employer Street Address'), Usage('S'), Position(50), Required(True)]
    
    n4: Annotated[L2100D_N4, Title('Member Employer City, State, Zip'), Usage('S'), Position(60), Syntax(['C0605']), Required(True)]


class L2100E_NM1(Segment):
    """Member School"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['M8'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['2'])]
    
    nm103: Annotated[D_1035, Title('School Name'), Usage('R'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Name First'), Usage('N'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Name Middle'), Usage('N'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Name Prefix'), Usage('N'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Name Suffix'), Usage('N'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('N'), Position(8)]
    
    nm109: Annotated[D_67, Title('Identification Code'), Usage('N'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100E_PER(Segment):
    """Member School Communications Numbers"""
    _segment_name = 'PER'
    
    per01: Annotated[D_366, Title('Contact Function Code'), Usage('R'), Position(1), Enumerated(*['SK'])]
    
    per02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    per03: Annotated[D_365, Title('Communication Number Qualifier'), Usage('R'), Position(3), Enumerated(*['EM', 'EX', 'FX', 'TE'])]
    
    per04: Annotated[D_364, Title('Communication Number'), Usage('R'), Position(4)]
    
    per05: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(5), Enumerated(*['EM', 'EX', 'FX', 'TE'])]
    
    per06: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(6)]
    
    per07: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(7), Enumerated(*['EM', 'EX', 'FX', 'TE'])]
    
    per08: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(8)]
    
    per09: Annotated[D_443, Title('Contact Inquiry Reference'), Usage('N'), Position(9)]


class L2100E_N3(Segment):
    """Member School Street Address"""
    _segment_name = 'N3'
    
    n301: Annotated[D_166, Title('School Address Line'), Usage('R'), Position(1)]
    
    n302: Annotated[D_166, Title('School Address Line'), Usage('S'), Position(2)]


class L2100E_N4(Segment):
    """Member School City, State, Zip"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('School City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('School State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('School Postal Zone or ZIP Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4), Enumerated(*country)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('N'), Position(5)]
    
    n406: Annotated[D_310, Title('Location Identifier'), Usage('N'), Position(6)]


class L2100E(Loop):
    
    nm1: Annotated[L2100E_NM1, Title('Member School'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    per: Annotated[L2100E_PER, Title('Member School Communications Numbers'), Usage('S'), Position(40), Syntax(['P0304', 'P0506', 'P0708']), Required(True)]
    
    n3: Annotated[L2100E_N3, Title('Member School Street Address'), Usage('S'), Position(50), Required(True)]
    
    n4: Annotated[L2100E_N4, Title('Member School City, State, Zip'), Usage('S'), Position(60), Syntax(['C0605']), Required(True)]


class L2100F_NM1(Segment):
    """Custodial Parent"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['S3'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1'])]
    
    nm103: Annotated[D_1035, Title('Custodial Parent Last Name'), Usage('R'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Custodial Parent First Name'), Usage('R'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Custodial Parent Middle Name'), Usage('S'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Custodial Parent Name Prefix'), Usage('S'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Custodial Parent Name Suffix'), Usage('S'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(8), Enumerated(*['34', 'ZZ'])]
    
    nm109: Annotated[D_67, Title('Custodial Parent Identifier'), Usage('S'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100F_PER(Segment):
    """Custodial Parent Communications Numbers"""
    _segment_name = 'PER'
    
    per01: Annotated[D_366, Title('Contact Function Code'), Usage('R'), Position(1), Enumerated(*['PQ'])]
    
    per02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    per03: Annotated[D_365, Title('Communication Number Qualifier'), Usage('R'), Position(3), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per04: Annotated[D_364, Title('Communication Number'), Usage('R'), Position(4)]
    
    per05: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(5), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per06: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(6)]
    
    per07: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(7), Enumerated(*['EM', 'EX', 'HP', 'TE', 'WP'])]
    
    per08: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(8)]
    
    per09: Annotated[D_443, Title('Contact Inquiry Reference'), Usage('N'), Position(9)]


class L2100F_N3(Segment):
    """Custodial Parent Street Address"""
    _segment_name = 'N3'
    
    n301: Annotated[D_166, Title('Custodial Parent Address Line'), Usage('R'), Position(1)]
    
    n302: Annotated[D_166, Title('Custodial Parent Address Line'), Usage('S'), Position(2)]


class L2100F_N4(Segment):
    """Custodial Parent City, State, Zip"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('Custodial Parent City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('Custodial Parent State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('Custodial Parent Postal Zone or ZIP Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4), Enumerated(*country)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('N'), Position(5)]
    
    n406: Annotated[D_310, Title('Location Identifier'), Usage('N'), Position(6)]


class L2100F(Loop):
    
    nm1: Annotated[L2100F_NM1, Title('Custodial Parent'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    per: Annotated[L2100F_PER, Title('Custodial Parent Communications Numbers'), Usage('S'), Position(40), Syntax(['P0304', 'P0506', 'P0708']), Required(True)]
    
    n3: Annotated[L2100F_N3, Title('Custodial Parent Street Address'), Usage('S'), Position(50), Required(True)]
    
    n4: Annotated[L2100F_N4, Title('Custodial Parent City, State, Zip'), Usage('S'), Position(60), Syntax(['C0605']), Required(True)]


class L2100G_NM1(Segment):
    """Responsible Person"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['E1', 'EI', 'EXS', 'GD', 'J6', 'QD'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1'])]
    
    nm103: Annotated[D_1035, Title('Responsible Party Last or Organization Name'), Usage('R'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Responsible Party First Name'), Usage('R'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Responsible Party Middle Name'), Usage('S'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Responsible Party Name Prefix'), Usage('S'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Responsible Party Suffix Name'), Usage('S'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(8), Enumerated(*['34', 'ZZ'])]
    
    nm109: Annotated[D_67, Title('Responsible Party Identifier'), Usage('S'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(10)]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2100G_PER(Segment):
    """Responsible Person Communications Numbers"""
    _segment_name = 'PER'
    
    per01: Annotated[D_366, Title('Contact Function Code'), Usage('R'), Position(1), Enumerated(*['RP'])]
    
    per02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    per03: Annotated[D_365, Title('Communication Number Qualifier'), Usage('R'), Position(3), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per04: Annotated[D_364, Title('Communication Number'), Usage('R'), Position(4)]
    
    per05: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(5), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per06: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(6)]
    
    per07: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(7), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per08: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(8)]
    
    per09: Annotated[D_443, Title('Contact Inquiry Reference'), Usage('N'), Position(9)]


class L2100G_N3(Segment):
    """Responsible Person Street Address"""
    _segment_name = 'N3'
    
    n301: Annotated[D_166, Title('Responsible Party Address Line'), Usage('R'), Position(1)]
    
    n302: Annotated[D_166, Title('Responsible Party Address Line'), Usage('S'), Position(2)]


class L2100G_N4(Segment):
    """Responsible Person City, State, Zip"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('Responsible Party City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('Responsible Party State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('Responsible Party Postal Zone or ZIP Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4), Enumerated(*country)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('N'), Position(5)]
    
    n406: Annotated[D_310, Title('Location Identifier'), Usage('N'), Position(6)]


class L2100G(Loop):
    
    nm1: Annotated[L2100G_NM1, Title('Responsible Person'), Usage('R'), Position(30), Syntax(['P0809', 'C1110']), Required(True)]
    
    per: Annotated[L2100G_PER, Title('Responsible Person Communications Numbers'), Usage('S'), Position(40), Syntax(['P0304', 'P0506', 'P0708']), Required(True)]
    
    n3: Annotated[L2100G_N3, Title('Responsible Person Street Address'), Usage('S'), Position(50), Required(True)]
    
    n4: Annotated[L2100G_N4, Title('Responsible Person City, State, Zip'), Usage('S'), Position(60), Syntax(['C0605']), Required(True)]


class L2200_DSB(Segment):
    """Disability Information"""
    _segment_name = 'DSB'
    
    dsb01: Annotated[D_1146, Title('Disability Type Code'), Usage('R'), Position(1), Enumerated(*['1', '2', '3', '4'])]
    
    dsb02: Annotated[D_380, Title('Quantity'), Usage('N'), Position(2)]
    
    dsb03: Annotated[D_1149, Title('Occupation Code'), Usage('N'), Position(3)]
    
    dsb04: Annotated[D_1154, Title('Work Intensity Code'), Usage('N'), Position(4)]
    
    dsb05: Annotated[D_1161, Title('Product Option Code'), Usage('N'), Position(5)]
    
    dsb06: Annotated[D_782, Title('Monetary Amount'), Usage('N'), Position(6)]
    
    dsb07: Annotated[D_235, Title('Product or Service ID Qualifier'), Usage('S'), Position(7), Enumerated(*['DX'])]
    
    dsb08: Annotated[D_1137, Title('Diagnosis Code'), Usage('S'), Position(8)]


class L2200_DTP(Segment):
    """Disability Eligibility Dates"""
    _segment_name = 'DTP'
    
    dtp01: Annotated[D_374, Title('Date Time Qualifier'), Usage('R'), Position(1), Enumerated(*['360', '361'])]
    
    dtp02: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(2), Enumerated(*['D8'])]
    
    dtp03: Annotated[D_1251, Title('Disability Eligibility Date'), Usage('R'), Position(3)]


class L2200(Loop):
    
    dsb: Annotated[L2200_DSB, Title('Disability Information'), Usage('R'), Position(200), Syntax(['P0708']), Required(True)]
    ItemDtp: TypeAlias = Annotated[L2200_DTP, Title('Disability Eligibility Dates'), Usage('S'), Position(210), Required(True)]
    dtp: Annotated[list[ItemDtp], MaxItems(2)]


class L2300_HD(Segment):
    """Health Coverage"""
    _segment_name = 'HD'
    
    hd01: Annotated[D_875, Title('Maintenance Type Code'), Usage('R'), Position(1), Enumerated(*['001', '002', '021', '024', '025', '026', '030', '032'])]
    
    hd02: Annotated[D_1203, Title('Maintenance Reason Code'), Usage('N'), Position(2)]
    
    hd03: Annotated[D_1205, Title('Insurance Line Code'), Usage('R'), Position(3), Enumerated(*['AG', 'AH', 'AJ', 'AK', 'HE', 'MM', 'UR', 'DCP', 'DEN', 'EPO', 'FAC', 'HE', 'HLT', 'HMO', 'LTC', 'LTD', 'MM', 'MOD', 'PDG', 'POS', 'PPO', 'PRA', 'STD', 'UR', 'VIS'])]
    
    hd04: Annotated[D_1204, Title('Plan Coverage Description'), Usage('S'), Position(4)]
    
    hd05: Annotated[D_1207, Title('Coverage Level Code'), Usage('S'), Position(5), Enumerated(*['CHD', 'DEP', 'E1D', 'E2D', 'E3D', 'E5D', 'E6D', 'E7D', 'E8D', 'E9D', 'ECH', 'EMP', 'ESP', 'FAM', 'IND', 'SPC', 'SPO', 'TWO'])]
    
    hd06: Annotated[D_609, Title('Count'), Usage('N'), Position(6)]
    
    hd07: Annotated[D_609, Title('Count'), Usage('N'), Position(7)]
    
    hd08: Annotated[D_1209, Title('Underwriting Decision Code'), Usage('N'), Position(8)]
    
    hd09: Annotated[D_1073, Title('Yes/No Condition or Response Code'), Usage('N'), Position(9)]
    
    hd10: Annotated[D_1211, Title('Drug House Code'), Usage('N'), Position(10)]
    
    hd11: Annotated[D_1073, Title('Yes/No Condition or Response Code'), Usage('N'), Position(11)]


class L2300_DTP(Segment):
    """Health Coverage Dates"""
    _segment_name = 'DTP'
    
    dtp01: Annotated[D_374, Title('Date Time Qualifier'), Usage('R'), Position(1), Enumerated(*['303', '348', '349', '543'])]
    
    dtp02: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(2), Enumerated(*['D8'])]
    
    dtp03: Annotated[D_1251, Title('Coverage Period'), Usage('R'), Position(3)]


class L2300_AMT(Segment):
    """Health Coverage Policy"""
    _segment_name = 'AMT'
    
    amt01: Annotated[D_522, Title('Amount Qualifier Code'), Usage('R'), Position(1), Enumerated(*['B9', 'C1', 'D2', 'P3'])]
    
    amt02: Annotated[D_782, Title('Contract Amount'), Usage('R'), Position(2)]
    
    amt03: Annotated[D_478, Title('Credit/Debit Flag Code'), Usage('N'), Position(3)]


class L2300_C040(Composite):
    """Reference Identifier"""
    pass


class L2300_REF(Segment):
    """Health Coverage Policy Number"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['17', '1L', 'ZZ'])]
    
    ref02: Annotated[D_127, Title('Insured Group or Policy Number'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class L2300_IDC(Segment):
    """Identification Card"""
    _segment_name = 'IDC'
    
    idc01: Annotated[D_1204, Title('Plan Coverage Description'), Usage('R'), Position(1)]
    
    idc02: Annotated[D_1215, Title('Identification Card Type Code'), Usage('R'), Position(2), Enumerated(*['D', 'H', 'P'])]
    
    idc03: Annotated[D_380, Title('Identification Card Count'), Usage('S'), Position(3)]
    
    idc04: Annotated[D_306, Title('Action Code'), Usage('S'), Position(4), Enumerated(*['1', '2', 'RX'])]


class L2310_LX(Segment):
    """Provider Information"""
    _segment_name = 'LX'
    
    lx01: Annotated[D_554, Title('Assigned Number'), Usage('R'), Position(1)]


class L2310_NM1(Segment):
    """Provider Name"""
    _segment_name = 'NM1'
    
    nm101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['3D', 'OD', 'P3', 'QA', 'QN', 'Y2'])]
    
    nm102: Annotated[D_1065, Title('Entity Type Qualifier'), Usage('R'), Position(2), Enumerated(*['1', '2'])]
    
    nm103: Annotated[D_1035, Title('Provider Last or Organization Name'), Usage('S'), Position(3)]
    
    nm104: Annotated[D_1036, Title('Provider First Name'), Usage('S'), Position(4)]
    
    nm105: Annotated[D_1037, Title('Provider Middle Name'), Usage('S'), Position(5)]
    
    nm106: Annotated[D_1038, Title('Provider Name Prefix'), Usage('S'), Position(6)]
    
    nm107: Annotated[D_1039, Title('Provider Name Suffix'), Usage('S'), Position(7)]
    
    nm108: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(8), Enumerated(*['34', 'FI', 'SV', 'XX'])]
    
    nm109: Annotated[D_67, Title('Provider Identifier'), Usage('S'), Position(9)]
    
    nm110: Annotated[D_706, Title('Entity Relationship Code'), Usage('R'), Position(10), Enumerated(*['25', '26', '72'])]
    
    nm111: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(11)]


class L2310_N4(Segment):
    """Provider City, State, Zip Code"""
    _segment_name = 'N4'
    
    n401: Annotated[D_19, Title('Member City Name'), Usage('R'), Position(1)]
    
    n402: Annotated[D_156, Title('Member State Code'), Usage('R'), Position(2), Enumerated(*states)]
    
    n403: Annotated[D_116, Title('Member Postal Zone or Zip Code'), Usage('R'), Position(3)]
    
    n404: Annotated[D_26, Title('Country Code'), Usage('S'), Position(4), Enumerated(*country)]
    
    n405: Annotated[D_309, Title('Location Qualifier'), Usage('S'), Position(5), Enumerated(*['60', 'CY', 'RJ'])]
    
    n406: Annotated[D_310, Title('Location Identification Code'), Usage('S'), Position(6)]


class L2310_PER(Segment):
    """Provider Communications Numbers"""
    _segment_name = 'PER'
    
    per01: Annotated[D_366, Title('Contact Function Code'), Usage('R'), Position(1), Enumerated(*['IC'])]
    
    per02: Annotated[D_93, Title('Name'), Usage('N'), Position(2)]
    
    per03: Annotated[D_365, Title('Communication Number Qualifier'), Usage('R'), Position(3), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per04: Annotated[D_364, Title('Communication Number'), Usage('R'), Position(4)]
    
    per05: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(5), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per06: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(6)]
    
    per07: Annotated[D_365, Title('Communication Number Qualifier'), Usage('S'), Position(7), Enumerated(*['EM', 'EX', 'FX', 'HP', 'TE', 'WP'])]
    
    per08: Annotated[D_364, Title('Communication Number'), Usage('S'), Position(8)]
    
    per09: Annotated[D_443, Title('Contact Inquiry Reference'), Usage('N'), Position(9)]


class L2310_PLA(Segment):
    """PCP Change Reason"""
    _segment_name = 'PLA'
    
    pla01: Annotated[D_306, Title('Action Code'), Usage('R'), Position(1), Enumerated(*['2'])]
    
    pla02: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(2), Enumerated(*['1P'])]
    
    pla03: Annotated[D_373, Title('Provider Effective Date'), Usage('R'), Position(3)]
    
    pla04: Annotated[D_337, Title('Time'), Usage('N'), Position(4)]
    
    pla05: Annotated[D_1203, Title('Maintenance Reason Code'), Usage('R'), Position(5), Enumerated(*['14', '22', '46', 'AA', 'AB', 'AC', 'AD', 'AE', 'AF', 'AG', 'AH', 'AI', 'AJ'])]


class L2310(Loop):
    
    lx: Annotated[L2310_LX, Title('Provider Information'), Usage('R'), Position(310), Required(True)]
    
    nm1: Annotated[L2310_NM1, Title('Provider Name'), Usage('R'), Position(320), Syntax(['P0809', 'C1110']), Required(True)]
    
    n4: Annotated[L2310_N4, Title('Provider City, State, Zip Code'), Usage('S'), Position(360), Syntax(['C0605']), Required(True)]
    ItemPer: TypeAlias = Annotated[L2310_PER, Title('Provider Communications Numbers'), Usage('S'), Position(370), Syntax(['P0304', 'P0506', 'P0708']), Required(True)]
    per: Annotated[list[ItemPer], MaxItems(2)]
    
    pla: Annotated[L2310_PLA, Title('PCP Change Reason'), Usage('S'), Position(395), Required(True)]


class L2320_COB(Segment):
    """Coordination of Benefits"""
    _segment_name = 'COB'
    
    cob01: Annotated[D_1138, Title('Payer Responsibility Sequence Number Code'), Usage('R'), Position(1), Enumerated(*['P', 'S', 'T', 'U'])]
    
    cob02: Annotated[D_127, Title('Insured Group or Policy Number'), Usage('S'), Position(2)]
    
    cob03: Annotated[D_1143, Title('Coordination of Benefits Code'), Usage('R'), Position(3), Enumerated(*['1', '5', '6'])]


class L2320_C040(Composite):
    """Reference Identifier"""
    pass


class L2320_REF(Segment):
    """Additional Coordination of Benefits Identifiers"""
    _segment_name = 'REF'
    
    ref01: Annotated[D_128, Title('Reference Identification Qualifier'), Usage('R'), Position(1), Enumerated(*['60', '6P', 'A6', 'SY', 'ZZ'])]
    
    ref02: Annotated[D_127, Title('Insured Group or Policy Number'), Usage('R'), Position(2)]
    
    ref03: Annotated[D_352, Title('Description'), Usage('N'), Position(3)]


class L2320_N1(Segment):
    """Other Insurance Company Name"""
    _segment_name = 'N1'
    
    n101: Annotated[D_98, Title('Entity Identifier Code'), Usage('R'), Position(1), Enumerated(*['IN'])]
    
    n102: Annotated[D_93, Title('Insurer Name'), Usage('S'), Position(2)]
    
    n103: Annotated[D_66, Title('Identification Code Qualifier'), Usage('S'), Position(3), Enumerated(*['FI', 'NI', 'XV'])]
    
    n104: Annotated[D_67, Title('Insured Group or Policy Number'), Usage('S'), Position(4)]
    
    n105: Annotated[D_706, Title('Entity Relationship Code'), Usage('N'), Position(5)]
    
    n106: Annotated[D_98, Title('Entity Identifier Code'), Usage('N'), Position(6)]


class L2320_DTP(Segment):
    """Coordination of Benefits Eligibility Dates"""
    _segment_name = 'DTP'
    
    dtp01: Annotated[D_374, Title('Date Time Qualifier'), Usage('R'), Position(1), Enumerated(*['344', '345'])]
    
    dtp02: Annotated[D_1250, Title('Date Time Period Format Qualifier'), Usage('R'), Position(2), Enumerated(*['D8'])]
    
    dtp03: Annotated[D_1251, Title('Coordination of Benefits Date'), Usage('R'), Position(3)]


class L2320(Loop):
    
    cob: Annotated[L2320_COB, Title('Coordination of Benefits'), Usage('R'), Position(400), Required(True)]
    ItemRef: TypeAlias = Annotated[L2320_REF, Title('Additional Coordination of Benefits Identifiers'), Usage('S'), Position(405), Syntax(['R0203']), Required(True)]
    ref: Annotated[list[ItemRef], MaxItems(5)]
    
    n1: Annotated[L2320_N1, Title('Other Insurance Company Name'), Usage('S'), Position(410), Syntax(['R0203', 'P0304']), Required(True)]
    ItemDtp: TypeAlias = Annotated[L2320_DTP, Title('Coordination of Benefits Eligibility Dates'), Usage('S'), Position(450), Required(True)]
    dtp: Annotated[list[ItemDtp], MaxItems(2)]


class L2300(Loop):
    
    hd: Annotated[L2300_HD, Title('Health Coverage'), Usage('R'), Position(260), Required(True)]
    ItemDtp: TypeAlias = Annotated[L2300_DTP, Title('Health Coverage Dates'), Usage('R'), Position(270), Required(True)]
    dtp: Annotated[list[ItemDtp], MaxItems(4)]
    ItemAmt: TypeAlias = Annotated[L2300_AMT, Title('Health Coverage Policy'), Usage('S'), Position(280), Required(True)]
    amt: Annotated[list[ItemAmt], MaxItems(4)]
    ItemRef: TypeAlias = Annotated[L2300_REF, Title('Health Coverage Policy Number'), Usage('S'), Position(290), Syntax(['R0203']), Required(True)]
    ref: Annotated[list[ItemRef], MaxItems(2)]
    ItemIdc: TypeAlias = Annotated[L2300_IDC, Title('Identification Card'), Usage('S'), Position(300), Required(True)]
    idc: Annotated[list[ItemIdc], MaxItems(10)]
    ItemL2310: TypeAlias = Annotated[L2310, Title('Provider Information'), Usage('S'), Position(310), Required(True)]
    l2310: Annotated[list[ItemL2310], MinItems(30)]
    ItemL2320: TypeAlias = Annotated[L2320, Title('Coordination of Benefits'), Usage('S'), Position(400), Required(True)]
    l2320: Annotated[list[ItemL2320], MinItems(5)]


class L2000(Loop):
    
    ins: Annotated[L2000_INS, Title('Member Level Detail'), Usage('R'), Position(10), Syntax(['P1112']), Required(True)]
    
    ref: Annotated[L2000_REF, Title('Subscriber Number'), Usage('R'), Position(20), Syntax(['R0203']), Required(True)]
    
    ref: Annotated[L2000_REF, Title('Member Policy Number'), Usage('S'), Position(20), Syntax(['R0203']), Required(True)]
    ItemRef: TypeAlias = Annotated[L2000_REF, Title('Member Identification Number'), Usage('S'), Position(20), Syntax(['R0203']), Required(True)]
    ref: Annotated[list[ItemRef], MaxItems(5)]
    
    ref: Annotated[L2000_REF, Title('Prior Coverage Months'), Usage('S'), Position(20), Syntax(['R0203']), Required(True)]
    ItemDtp: TypeAlias = Annotated[L2000_DTP, Title('Member Level Dates'), Usage('S'), Position(25), Required(True)]
    dtp: Annotated[list[ItemDtp], MaxItems(20)]
    ItemL2100A: TypeAlias = Annotated[L2100A, Title('Member Name'), Usage('R'), Position(30), Required(True)]
    l2100a: Annotated[list[ItemL2100A], MinItems(1)]
    ItemL2100B: TypeAlias = Annotated[L2100B, Title('Incorrect Member Name'), Usage('S'), Position(30), Required(True)]
    l2100b: Annotated[list[ItemL2100B], MinItems(1)]
    ItemL2100C: TypeAlias = Annotated[L2100C, Title('Member Mailing Address'), Usage('S'), Position(30), Required(True)]
    l2100c: Annotated[list[ItemL2100C], MinItems(1)]
    ItemL2100D: TypeAlias = Annotated[L2100D, Title('Member Employer'), Usage('S'), Position(30), Required(True)]
    l2100d: Annotated[list[ItemL2100D], MinItems(3)]
    ItemL2100E: TypeAlias = Annotated[L2100E, Title('Member School'), Usage('S'), Position(30), Required(True)]
    l2100e: Annotated[list[ItemL2100E], MinItems(3)]
    ItemL2100F: TypeAlias = Annotated[L2100F, Title('Custodial Parent'), Usage('S'), Position(30), Required(True)]
    l2100f: Annotated[list[ItemL2100F], MinItems(1)]
    ItemL2100G: TypeAlias = Annotated[L2100G, Title('Responsible Person'), Usage('S'), Position(30), Required(True)]
    l2100g: Annotated[list[ItemL2100G], MinItems(1)]
    ItemL2200: TypeAlias = Annotated[L2200, Title('Disability Information'), Usage('S'), Position(200), Required(True)]
    l2200: Annotated[list[ItemL2200], MinItems(1)]
    ItemL2300: TypeAlias = Annotated[L2300, Title('Health Coverage'), Usage('S'), Position(260), Required(True)]
    l2300: Annotated[list[ItemL2300], MinItems(99)]


class DETAIL(Loop):
    ItemL2000: TypeAlias = Annotated[L2000, Title('Member Level Detail'), Usage('R'), Position(10), Required(True)]
    l2000: Annotated[list[ItemL2000], MinItems(1)]


class FOOTER(Loop):
    pass


class ST_LOOP_SE(Segment):
    """Transaction Set Trailer"""
    _segment_name = 'SE'
    
    se01: Annotated[D_96, Title('Transaction Segment Count'), Usage('R'), Position(1)]
    
    se02: Annotated[D_329, Title('Transaction Set Control Number'), Usage('R'), Position(2)]


class ST_LOOP(Loop):
    
    st: Annotated[ST_LOOP_ST, Title('Transaction Set Header'), Usage('R'), Position(5), Required(True)]
    ItemHeader: TypeAlias = Annotated[HEADER, Title('Table 1 - Header'), Usage('R'), Position(10), Required(True)]
    header: Annotated[list[ItemHeader], MinItems(1)]
    ItemDetail: TypeAlias = Annotated[DETAIL, Title('Table 2 - Detail'), Usage('S'), Position(20), Required(True)]
    detail: Annotated[list[ItemDetail], MinItems(1)]
    ItemFooter: TypeAlias = Annotated[FOOTER, Title('Table 3 - Footer'), Usage('N'), Position(30)]
    footer: Annotated[list[ItemFooter], MinItems(1)]
    
    se: Annotated[ST_LOOP_SE, Title('Transaction Set Trailer'), Usage('R'), Position(690), Required(True)]


class GS_LOOP_GE(Segment):
    """Functional Group Trailer"""
    _segment_name = 'GE'
    
    ge01: Annotated[D_97, Title('Number of Transaction Sets Included'), Usage('R'), Position(1)]
    
    ge02: Annotated[D_28, Title('Group Control Number'), Usage('R'), Position(2)]


class GS_LOOP(Loop):
    
    gs: Annotated[GS_LOOP_GS, Title('Functional Group Header'), Usage('R'), Position(10), Required(True)]
    ItemSt_Loop: TypeAlias = Annotated[ST_LOOP, Title('Transaction Set Header'), Usage('R'), Position(20), Required(True)]
    st_loop: Annotated[list[ItemSt_Loop], MinItems(1)]
    
    ge: Annotated[GS_LOOP_GE, Title('Functional Group Trailer'), Usage('R'), Position(30), Required(True)]


class ISA_LOOP_TA1(Segment):
    """Interchange Acknowledgement"""
    _segment_name = 'TA1'
    
    ta101: Annotated[I12, Title('Interchange Control Number'), Usage('R'), Position(1)]
    
    ta102: Annotated[I08, Title('Interchange Date'), Usage('R'), Position(2)]
    
    ta103: Annotated[I09, Title('Interchange Time'), Usage('R'), Position(3)]
    
    ta104: Annotated[I17, Title('Interchange Acknowledgement Code'), Usage('R'), Position(4), Enumerated(*['A', 'E', 'R'])]
    
    ta105: Annotated[I18, Title('Interchange Note Code'), Usage('R'), Position(5), Enumerated(*['000', '001', '002', '003', '004', '005', '006', '007', '008', '009', '010', '011', '012', '013', '014', '015', '016', '017', '018', '019', '020', '021', '022', '023', '024', '025', '026', '027', '028', '029', '030', '031'])]


class ISA_LOOP_IEA(Segment):
    """Interchange Control Trailer"""
    _segment_name = 'IEA'
    
    iea01: Annotated[I16, Title('Number of Included Functional Groups'), Usage('R'), Position(1)]
    
    iea02: Annotated[I12, Title('Interchange Control Number'), Usage('R'), Position(2)]


class ISA_LOOP(Loop):
    
    isa: Annotated[ISA_LOOP_ISA, Title('Interchange Control Header'), Usage('R'), Position(10), Required(True)]
    ItemGs_Loop: TypeAlias = Annotated[GS_LOOP, Title('Functional Group Header'), Usage('R'), Position(20), Required(True)]
    gs_loop: Annotated[list[ItemGs_Loop], MinItems(1)]
    
    ta1: Annotated[ISA_LOOP_TA1, Title('Interchange Acknowledgement'), Usage('S'), Position(20), Required(True)]
    
    iea: Annotated[ISA_LOOP_IEA, Title('Interchange Control Trailer'), Usage('R'), Position(30), Required(True)]


class MSG834(Message):
    """HIPAA Benefit Enrollment and Maintenance X095A1-834"""
    ItemIsa_Loop: TypeAlias = Annotated[ISA_LOOP, Title('Interchange Control Header'), Usage('R'), Position(1), Required(True)]
    isa_loop: Annotated[list[ItemIsa_Loop], MinItems(1)]



"""
The ``xml_extract`` application converts the schema from the PyX12 project's XML-based schema definitions
into Python class definitions using Annotated types.


SYNOPSIS
===========

::

    python tools/xml_extract.py

DESCRIPTION
============

This application creates subclasses of :py:class:`x12.base.Composite`, :py:class:`x12.base.Segment`, :py:class:`x12.base.Loop`, and :py:class:`x12.base.Message` definitions.
This writes a module file with the class definitions for each message.

It also creates a :py:mod:`x12.common` module with base
types and reused type definitions.


OPTIONS
=======

Currently, there are no command-line options. The ``__main__`` section sets
two global variables::

:base:

    This has default value of ``Path.home() / "github" / "pyx12" / "pyx12" / "map"``
    It's the location of the GitHub checked out of the PyX12 project.

:output:
    This is the format of the output. The default  value is ``"python"``.
        Other choices include ``"schema-python"`` and ``"json"``.

The results are written to the ``../x12`` or the ``../json`` directory.
For python and schema-python options, the output goes to ``../x12``.
The ``"json"`` output is written to ``../x12``.

EXIT STATUS
============

:0:
    if it worked.

:2:
    if it raises an exception. This is often because of unparsable XML input or unwritable output directories.


ENVIRONMENT
===========

TBD


FILES
=======

See https://github.com/azoner/pyx12

See https://github.com/azoner/pyx12/tree/master/pyx12/map for the message definitions.

Several sources of data are integrated to create the full schema.

1. The :file:`nnn.nnnn.xnnnn...xml` files. Example :file:`270.4010.X092.A1.xml`. These have the message definitions.

2. The :file:`codes.xml` file, and :file:`detaele.xml` file with additional data to support the data type definitions.

3. The :file:`x12.control.00401.xml` and :file:`x12.control.00501.xml` with core structural information
   used to map a message to the specific version of a message definition.

It's not perfectly clear that the :file:`maps.xml` file is *required*, but it seems useful for creating
a declarative set of X12 messages.

NOTES
==========

X12 Schema
-----------

The X12 schema definitions created here depend on an ``x12.base`` module to provide
the superclass definitions of Element, Composite, Segment, Loop, and Message.

This app writes a collection of message-specific modules
with class definitions for all of the elements, etc., that are part of that message.
Reuse is not considered at this point.
In many cases, the Control segment definitions (ISA, IEA, GS, GE, etc.) are reusable.
They're repeated in each module, however.

Structure and Naming
---------------------

One complication here is that within a given message definition,
the ``Element``, ``Segment``, ``Loop``, and ``Message`` structure is unique, but subsidiary structure names are reused.
This means the message -- as a whole -- is a kind of namespace around the structural elements.
This means Segment names are qualified by the Loop that contains the Segment.

The idea of exposing a Segment class as if it is a dataclass-like record
is made challenging by way Segments are reused in distinct Loops.

Type Definitions
-----------------

The Element definitions, used to define the atomic elements of Segments and Composites, are heavily reused.
These become a module of common definitions.

The X12 Data Type attribute is defines the base Python type.
These include numerous string types (``AN``, ``ID``),
integer types (``N``), float types (``R``), Decimal types (``Nx``),
datetime.date (``DT``) and datetime.time (``TM``).

There are 3 variations on DT formatting:

-   ``<data_ele ele_num="I08" data_type="DT" min_len="6" max_len="6" name="I08"/>``

-   ``<data_ele ele_num="373" data_type="DT" min_len="8" max_len="8" name="Date"/>``

-   ``<data_ele ele_num="29" data_type="DT" min_len="6" max_len="8" name="Date"/>``

This means there are three underlying data element definitions

::

    DT: TypeAlias = datetime.date

    I08: TypeAlias = Annotated[DT, MinLen(6), MaxLen(6)]
    D_373: TypeAlias = Annotated[DT, MinLen(8), MaxLen(8)]
    D_29: TypeAlias = Annotated[DT, MinLen(6), MaxLen(8)]

The Code definitions, similarly, are part of the module of schema details.
These are enumerations of all valid code values.
The raw values are in the common module. From these an :py:class:`x12.annotations.Enumerated`
annotation is built.

In a JSON Schema, a "$ref" path of the form "#/$common/some_code" or "#/$common/some_element_schema",
can provide a reference to the reusable data element definitions.
This parallels the Python ``x12.common`` module which has these definitions.

Using "#" implies  the "$common" section is part of each message's overall JSONSchema definition.
It could be replaced with an external reference using a full URI, which might be more
sensible in the long run.

Type Names and Stuctures
------------------------

There are two pools of type definitions in ``dataele.xml``

-   ``Ixx`` names, part of the reused ISA segment.

-   ``nnn`` numbers, used everywhere else. We rename these to ``D_nnn`` to provide a Pythonic name.

There are two pools of code definitions:

-   Common codesets defined in the ``codes.xml``.  (There are 9 of these.)

-   Data-element specific codesets defined within the element of a message.

A data type, then, has the following definition aspects:

-   Base ``data_type_code`` ("AN", "ID", "DT", "TM", etc.)
    These are -- generally -- mappings to a Python type.
    For the ``Nx``, it includes the decimal scale factor.
    For ``DT`` and ``TM``, a length is required, also.

-   Annotations with common definition details (minLength, maxLength, title) from ``common.xml``.
    A "format" value is implied by the length for ``DT``, ``TM``, ``Nx``, ``N``, and ``R``.

-   Enumerated literals come either from a ``codes.xml`` or from the element definition.


Type Extensions
---------------

Some messages will have two AN fields. One with a type and the second with a value.
This extra type detail is one of the following:

-   D8 (yyyymmdd),

-   D6 (yymmdd),

-   RD8 (yyyymmdd-yyyymmdd),

-   TM (hhmm)

These are types **outside** the X12 definitions.
They're application-specific.

X12 Message Version Mapping
---------------------------

The overall "mapping" from source text to message is a separate capability
This can use ``x12.control.*`` module generic definitions to locate the correct module to process a message.

The base mapping superclass permits a subclass to provide a rule
for disambiguating messages.

A mapping search class could use the collection of mapping objects
to locate the module that can parse a message.

.. todo:: This version identification is not implemented

"""
import abc
from collections import deque
from collections.abc import Iterator
from contextlib import redirect_stdout
from dataclasses import dataclass, asdict, field
import datetime
from decimal import Decimal
import logging
from pathlib import Path
from pprint import pprint, pformat
import sys
from textwrap import indent
from typing import (
    Union, TypeAlias, Any, TextIO,
    Self, cast, Deque, Type
)
from xml.etree import ElementTree as XML


logger = logging.getLogger("xml_extract")

optional_text = lambda tag, name: (
    sub_tag.text if (sub_tag := tag.find(name)) is not None else ""
)

JSONSchema: TypeAlias = dict[str, Any] | list[Any] | str

@dataclass
class Codeset:
    """
    An enumerated set of values used to define an Element.

    A JSONSchema ``{"type": "string", "enum": []}`` schema.

    No Annotation, these are ``name = [values]`` constructs in Python.
    """
    id: str
    data_ele: str
    source: str
    values: list[str]

    @property
    def label(self):
        return self.id

    def jsonschema(self) -> JSONSchema:
        return self.values


def code_reader(source: TextIO) -> Iterator[Codeset]:
    """
    See https://github.com/azoner/pyx12/blob/master/pyx12/map/codes.rng

    The RelaxNG schema doesn't exactly match the XML.

    The schema is approximately this
    ::

        schema:
            description: codesets
            type: array
            items:
                description: codeset
                type: object
                properties:
                    id:
                        type: string
                        maxLength: 20
                    name:
                        type: string
                        maxLength: 100
                    data_ele:
                        type: string
                        maxLength: 6
                    source:
                        type: string
                    version:
                        type: array
                        items:
                            description: version
                            type: object
                            properties:
                                id:
                                    type: string
                                    maxLength: 20
                                eff_dte:
                                    type: stringh
                                code:
                                    type: array
                                    items:
                                        type: str
                                        maxLength: 6
    """
    document = XML.parse(source)
    codesets = document.getroot()
    for codeset_tag in codesets.findall('codeset'):
        id = cast(XML.Element, codeset_tag.find('id')).text
        data_ele = optional_text(codeset_tag, 'data_ele')
        source = optional_text(codeset_tag, 'source')
        version_tag = cast(XML.Element, codeset_tag.find('version'))
        values = [
            code.text for code in version_tag.findall('code')
        ]
        codeset = Codeset(
            id=cast(str, id),
            data_ele=data_ele,
            source=cast(str, source),
            values=cast(list[str], values),
        )
        logger.debug(codeset)
        yield codeset


X12_TYPE_SCHEMA: dict[str, JSONSchema] = {
    "AN": {"type": "string"}, # AlphaNumeric
    "B": {"type": "string"},  # Binary?
    "DT": {"type": "string"},  # date, format based on length, "format": "\\d{6,8}"
    "ID": {"type": "string"}, # Identifier, numeric string.
    "R": {"type": "number"},  # Real
    "TM": {"type": "string"},  # Time. format "format": "\\d{4,6}"
    "N": {"type": "integer"},  # Number (integer?)
    # These have no proper JSONSchema representation, the underlying type is "Decimal" with a scale.
    "N0": {"type": "number", "scale": 0}, # 0 decimal places.
    "N1": {"type": "number", "scale": 1},
    "N2": {"type": "number", "scale": 2}, # 2 decimal places, i.e. currency.
    "N3": {"type": "number", "scale": 3},
    "N4": {"type": "number", "scale": 4},
    "N5": {"type": "number", "scale": 5},
    "N6": {"type": "number", "scale": 6},
    "N7": {"type": "number", "scale": 7},
    "N8": {"type": "number", "scale": 8},
    "N9": {"type": "number", "scale": 9},
}

X12_TYPE_ANNOTATION: dict[str, str] = {
    "AN": "str",  # AlphaNumeric
    "B": "str",  # Binary?
    "DT": "datetime.date",  # Annotation will add MinLen(6-8), MaxLen(6-8), Format(r"\d{6}{8}") details
    "ID": "str",  # Identifier, often (but not always) a numeric string.
    "R": "float",  # Real
    "TM": "datetime.time",  # Annotation will add MinLen(4-6), MaxLen(4-6), Format(r"\d{4}{6}") details
    "N": "int",  # Number (integer?)
    "N0": "Annotated[Decimal, Scale(0)]",  # 0 decimal places
    "N1": "Annotated[Decimal, Scale(1)]",
    "N2": "Annotated[Decimal, Scale(2)]",  # 2 decimal places, i.e. currency.
    "N3": "Annotated[Decimal, Scale(3)]",
    "N4": "Annotated[Decimal, Scale(4)]",
    "N5": "Annotated[Decimal, Scale(5)]",
    "N6": "Annotated[Decimal, Scale(6)]",
    "N7": "Annotated[Decimal, Scale(7)]",
    "N8": "Annotated[Decimal, Scale(8)]",
    "N9": "Annotated[Decimal, Scale(9)]",
}


@dataclass
class DataElement:
    """
    A data element description, often reused by multiple Elements.

    A JSONSchema {"type": ..., "minLength": ..., "maxLength": ..., "description": ...} schema
    """
    ele_num: str
    data_type: str  # Base type
    min_len: str | None
    max_len: str | None
    name: str

    @property
    def class_name(self) -> str:
        return f'{pythonify(self.ele_num, "D_")}'

    @property
    def int_min_len(self) -> int | None:
        if self.min_len is not None:
            return int(self.min_len)
        return None

    @property
    def int_max_len(self) -> int | None:
        if self.max_len is not None:
            return int(self.max_len)
        return None

    def jsonschema(self) -> JSONSchema:
        schema = cast(dict[str, Any], X12_TYPE_SCHEMA[self.data_type])
        schema |= {
            "title": self.name,
            "data_type_code": self.data_type,
        }
        if self.min_len is not None:
            schema["minLength"] = int(self.min_len)
        if self.max_len is not None:
            schema["maxLength"] = int(self.max_len)
        # additional special cases
        if self.data_type in {"DT", "TM"}:
            if self.min_len == self.max_len:
                schema["format"] = f"\\d{{{self.min_len}}}"
            else:
                schema["format"] = f"\\d{{{self.min_len},{self.max_len}}}"
        return schema

    def annotation(self) -> tuple[str, str]:
        base = X12_TYPE_ANNOTATION[self.data_type]
        extensions = [base, f"Title({self.name!r})"]
        if self.min_len is not None:
            extensions.append(f"MinLen({int(self.min_len)})")
        if self.max_len is not None:
            extensions.append(f"MaxLen({int(self.max_len)})")
        if self.data_type in {"DT", "TM"}:
            if self.min_len == self.max_len:
                extensions.append(f'Format("\\d{self.min_len}")')
            else:
                extensions.append(f'Format("\\d{self.min_len,self.max_len}")')
        return (
            f"{self.class_name}: TypeAlias = Annotated[{', '.join(extensions)}]",
            ""
        )

def element_reader(source: TextIO) -> Iterator[DataElement]:
    """
    See https://github.com/azoner/pyx12/blob/master/pyx12/map/codes.rng

    The RelaxNG schema seems to match the XML.
    Note that the schema is all attributes, not actual tags.

    The schema is approximately this
    ::

        description: data_elements
        type: array
        items:
            description: data_ele
            type: object
            properties:
                ele_num:
                    type: string
                    max-length: 4
                data_type:
                    type: string
                    enum: ["AN", "B", "DT", "ID", "R", "TM", "N", "N0", "N1", "N2",  "N3", "N4", "N5",  "N6", "N7", "N8",
                    "N9"]
                min_len:
                    type: number
                    maximum: 16384
                max_len:
                    type: number
                    maximum: 16384
                name:
                    type: string
    """
    document = XML.parse(source)
    data_elements = document.getroot()
    for data_ele_tag in data_elements.findall('data_ele'):
        ele_num = data_ele_tag.attrib.get('ele_num', '')
        data_type = data_ele_tag.attrib.get('data_type', '')
        min_len = data_ele_tag.attrib.get('min_len', '')
        max_len = data_ele_tag.attrib.get('max_len', '')
        name = data_ele_tag.attrib.get('name', '')
        assert data_type in ["AN", "B", "DT", "ID", "R", "TM", "N", "N0", "N1", "N2",  "N3", "N4", "N5",  "N6", "N7", "N8", "N9"]
        data_ele = DataElement(
            ele_num=ele_num,
            data_type=data_type,
            min_len=min_len,
            max_len=max_len,
            name=name,
        )
        logger.debug(data_ele)
        yield data_ele


@dataclass
class SchemaMap:
    """
    Rules for mapping a given message to a specific versioned type.
    """
    icvn: str  # Interchange Control Version Number (ISA12)
    vriic: str  # Version / Release / Industry Identifier Code (GS08)
    fic: str  # Functional Identifier Code (GS01)
    abbr: str
    definition: str


def map_reader(source: TextIO) -> Iterator[SchemaMap]:
    """
    Mapping of document to schema version.

    See https://github.com/azoner/pyx12/blob/master/pyx12/map/maps.xsd

    Attributes used to distinguish the document schema

    - icvn - Interchange Control Version Number (ISA12)
    - fic - Functional Identifier Code (GS01)
    - vriic - Version / Release / Industry Identifier Code (GS08)
    """
    document = XML.parse(source)
    maps_tag = document.getroot()
    for version_tag in maps_tag.findall("version"):
        icvn = version_tag.attrib.get("icvn")
        for map_tag in version_tag.findall("map"):
            vriic = map_tag.attrib.get("vriic")
            fic = map_tag.attrib.get("fic")
            abbr = map_tag.attrib.get("abbr")
            definition = map_tag.text
            map = SchemaMap(
                icvn=cast(str, icvn),
                vriic=cast(str, vriic),
                fic=cast(str, fic),
                abbr=cast(str, abbr),
                definition=cast(str, definition),
            )
            logger.info(map)
            yield map


def pythonify(x12_name: str, prefix="_") -> str:
    revised = x12_name.replace("-", "_").replace(".", "_")
    if revised[0] in "0123456789":
        revised = prefix + revised
    return revised


@dataclass
class Element:
    """
    Atomic Element.

    A JSONSchema property value: "name": {"type": ..., etc.} piece of a Segment or Composite

    Yes, codes are locally defined codes, where code_set is an imported set of codes.
    They're both the same JSONSchema enum attribute.

    Usage="R" (required) percolates up to the containing Composite or Segment.
    """
    xid: str
    data_ele: str
    name: str
    usage: str
    seq: str
    refdes: str  # Not used
    repeat: str
    codes: list[str]
    external: str
    common_def: str  # DataElement or Codeset identifier
    loop: str  # Containing Loop to make a class name unique

    @property
    def attr_name(self) -> str:
        return pythonify(self.xid, "E").lower()

    @property
    def class_name(self) -> str:
        return f"{self.loop}_{pythonify(self.xid, 'E')}"

    def jsonschema(self) -> JSONSchema:
        schema: dict[str, Any] = {
            "title": self.name,
            "usage": self.usage,  # Required, Situational, Not-Used
            "description": f"xid={self.xid} data_ele={self.data_ele}",
            "position": int(self.seq),
        }
        # Can have both common type def & locally-defined codes.
        # This is a "allOf" merge between {$ref: common_def} type and {"enum": ...} for codes.
        if self.codes and not self.common_def:
            schema["type"] = "string"
            schema["enum"] = self.codes
        elif self.common_def and not self.codes:
            schema["type"] = {"$ref": f"#/$common/{self.common_def}"}
        elif self.common_def and (self.codes or self.external):
            schema["type"] = {
                "allOf": [
                    {"$ref": f"#/$common/{self.common_def}"},
                    {"enum": self.codes or f"#/$codes/{self.external}"}
                ]
            }
        else:
            raise RuntimeError(f"no type information in {self}")
        if self.repeat:
            schema = {
                "type": "array",
                "items": schema,
                "maxItems": int(self.repeat)
            }
        return schema

    def annotation(self) -> tuple[str, str]:
        """
        Some Segment or Composite will have

        ::

            name: Annotated[DT, Title("Item"), Usage("R"), Sequence("1")]

        The ``DT`` type alias is ``Annotated[datetime.date, ...]``.

        This flattens nicely to ``datetime.date`` with details.

        Or

        ::

            ItemName: TypeAlias = Annotated[DT, title("Item"), ...]
            name: Annotated[list[ItemName], MaxItems(4)]

        The list -- as a whole -- and the items both have annotations.
        The over-all ``list[]`` has a ``MaxItems()`` annotation.
        The Element within the list has the detailed annotations for this element.
        """
        base = [pythonify(self.data_ele, 'D_'), f"Title({self.name!r})",  f"Usage({self.usage!r})", f"Position({int(self.seq)})"]
        if self.codes:
            base.append(f"Enumerated(*{self.codes})")
        # if self.common_def:
        #     base.append(f"Enumerated(*data_element[{self.common_def}])")
        if self.external:
            base.append(f"Enumerated(*{self.external})")
        type_str = f"Annotated[{', '.join(base)}]"
        if self.repeat:
            annotations = (
                f"Item{self.attr_name.title()}: TypeAlias = {type_str}",
                f"{self.attr_name}: Annotated[list[Item{self.attr_name.title()}], MaxItems({int(self.repeat)})]"
            )
        else:
            annotations = (
                "",
                f"{self.attr_name}: {type_str}"
            )
        return annotations

@dataclass
class Composite:
    """
    Composite of multiple Elements; an inline segment-like collection of Elements.

    A JSONSchema {"type": "object": "properties": {...}} schema

    From a JSONSchema perspective, a composite isn't flattened into a sequence
    of attributes of the containing Segment.
    The composite is described as if it's a sub-segment.
    """
    xid: str
    data_ele: str
    name: str
    usage: str  # Required, Situational, Not-Used
    seq: str
    refdes: str  # When used, this seems to be the proper XID
    syntax: list[str]
    repeat: str
    elements: list[Element]
    loop: str  # Containing Loop to make a class name unique

    @property
    def attr_name(self) -> str:
        return pythonify(self.xid or self.refdes or self.data_ele, "C").lower()

    @property
    def class_name(self) -> str:
        return f"{self.loop}_{pythonify(self.xid or self.refdes or self.data_ele, 'C')}"

    @property
    def required(self):
        return [elt.attr_name for elt in self.elements if elt.usage == "R"]

    def jsonschema(self) -> JSONSchema:
        schema = {
            "title": self.name,
            "usage": self.usage,  # Required, Situational, Not-Used
            "description": f"xid={self.xid} name={self.name} refdes={self.refdes} data_ele={self.data_ele}",
            "sequence": int(self.seq),
            "syntax": self.syntax,
        }
        if self.elements:
            schema["type"] = "object"
            schema["properties"] = {
                elt.attr_name: elt.jsonschema()
                for elt in self.elements
            }
        if req := self.required:
            schema["required"] = req
        if self.repeat:
            schema = {
                "type": "array",
                "items": schema,
                "maxItems": int(self.repeat)
            }
        return schema

    def annotation(self) -> tuple[str, str]:
        base = [self.class_name or pythonify(self.data_ele, 'D_'), f"Title({self.name!r})", f"Usage({self.usage!r})", f"Position({int(self.seq)})"]
        if self.required:
            base.append("Required(True)")
        type_str = f"Annotated[{', '.join(base)}]"
        if self.repeat:
            annotations = (
                f"Item{self.attr_name.title()}: TypeAlias = {type_str}",
                f"{self.attr_name}: Annotated[list[Item{self.attr_name.title()}], MaxItems({int(self.repeat)})]"
            )
        else:
            annotations = (
                "",
                f"{self.attr_name}: {type_str}"
            )
        return annotations

@dataclass
class Segment:
    """
    A collection of Elements and Composites.

    A JSONSchema {"type": "object": "properties": {...}} schema
    """
    xid: str
    name: str
    end_tag: str  # Used rarely
    usage: str  # Required, Situational, Not-Used
    pos: str
    max_use: str
    syntax: list[str]
    children: list[Element | Composite]
    loop: str  # Containing Loop to make a class name unique

    @property
    def attr_name(self) -> str:
        return pythonify(self.xid, "S").lower()

    @property
    def class_name(self) -> str:
        return f"{self.loop}_{pythonify(self.xid, 'S')}"

    @property
    def required(self) -> list[str]:
        return [elt.attr_name for elt in self.children if elt.usage == "R"]

    @property
    def repeat(self) -> bool:
        return bool(self.max_use) and self.max_use != "1"

    def jsonschema(self) -> JSONSchema:
        schema: dict[str, Any] = {
            "title": self.name,
            "usage": self.usage,  # Required, Situational, Not-Used
            "description": f"xid={self.xid} name={self.name}",
            "position": int(self.pos),
        }
        if self.syntax:
            schema["syntax"] = self.syntax
        if self.end_tag:
            schema["endTag"] = self.end_tag
        if self.children:
            schema["type"] = "object"
            schema["properties"] = {"xid": {"literal": self.xid}}
            for elt in self.children:
                # if elt.usage == "N":
                #     continue
                schema["properties"][elt.attr_name] = {"$ref": f"#/$elements/{elt.class_name}"}  # elt.jsonschema()
        if req := self.required:
            schema["required"] = req
        if self.max_use and self.max_use != "1":
            schema = {
                "type": "array",
                "items": schema,
            }
            if self.max_use != ">1":
                schema['maxItems'] = int(self.max_use)
        return schema

    def annotation(self) -> tuple[str, str]:
        """
        Some Message or Loop will have

        ::

            segment: Annotated[Loop_Seg_Class, Title("Segnment"), Usage("R"), Sequence("1")]

        This flattens nicely to ``Loop_Seg_Class`` with details.

        Or

        ::

            ItemSegment: TypeAlias = Annotated[Loop_Seg_Class, title("Segnment"), ...]
            segment: Annotated[list[ItemSegment], MaxItems(4)]

        The list -- as a whole -- and the items both have annotations.
        List annotations are small (MaxItems).
        The Segment in the list has more complex annotations.
        """
        # While this works, mypy can't figure it out.
        base = [self.class_name, f"Title({self.name!r})", f"Usage({self.usage!r})", f"Position({int(self.pos)})"]
        if self.syntax:
            base.append(f"Syntax({self.syntax})")
        if self.required:
            base.append("Required(True)")
        type_str = f"Annotated[{', '.join(base)}]"
        if self.repeat:
            if self.max_use.startswith(">"):
                min_max = f"MinItems({int(self.max_use[1:])})"
            else:
                min_max = f"MaxItems({int(self.max_use)})"
            annotations = (
                f"Item{self.attr_name.title()}: TypeAlias = {type_str}",
                f"{self.attr_name}: Annotated[list[Item{self.attr_name.title()}], {min_max}]"
            )
        else:
            annotations = (
                "",
                f"{self.attr_name}: {type_str}"
            )
        return annotations

@dataclass
class Loop:
    """
    A higher-level collection of Loops and Segments.

    A JSONSchema {"type": "object": "properties": {...}} schema
    """
    xid: str
    name: str
    usage: str  # Required, Situational, Not-Used
    pos: str
    repeat: str
    type: str
    children: list[Union[Segment, "Loop"]]

    @property
    def attr_name(self) -> str:
        return pythonify(self.xid, "L").lower()

    @property
    def class_name(self) -> str:
        return pythonify(self.xid, "L")

    @property
    def required(self) -> list[str]:
        return [elt.attr_name for elt in self.children if elt.usage == "R"]

    def jsonschema(self) -> JSONSchema:
        schema: dict[str, Any] = {
            "title": self.name,
            "usage": self.usage,  # Required, Situational, Not-Used
            "description": f"xid={self.xid} name={self.name} type={self.type}",
            "position": int(self.pos),
        }
        if self.children:
            schema["type"] = "object"
            schema["properties"] = {
                elt.attr_name: {"$ref": f"#/$segments/{elt.class_name}"}  # elt.jsonschema()
                for elt in self.children
                # if elt.usage != "N"
            }
        if req := self.required:
            schema["required"] = req
        if self.repeat:
            schema = {
                "type": "array",
                "items": schema,
            }
            if self.repeat != ">1":
                schema['maxItems'] = int(self.repeat)
        return schema

    def annotation(self) -> tuple[str, str]:
        base = [self.class_name, f"Title({self.name!r})", f"Usage({self.usage!r})", f"Position({int(self.pos)})"]
        if self.required:
            base.append("Required(True)")
        type_str = f"Annotated[{', '.join(base)}]"
        if self.repeat:
            if self.repeat.startswith('>'):
                min_items = f"MinItems({int(self.repeat[1:])})"
            else:
                min_items = f"MinItems({int(self.repeat)})"
            annotations = (
                f"Item{self.attr_name.title()}: TypeAlias = {type_str}",
                f"{self.attr_name}: Annotated[list[Item{self.class_name.title()}], {min_items}]"
            )
        else:
            annotations = (
                "",
                f"{self.attr_name}: {type_str}"
            )
        return annotations

@dataclass
class Message:
    """
    A Complete Message.

    No annotation for a message as a whole,
    it's a collection of Loops, wrapped
    in a ``class`` statement.
    """
    xid: str
    name: str
    loops: list[Loop]

    @property
    def class_name(self) -> str:
        return pythonify(self.xid, "MSG")

    @property
    def required(self) -> list[str]:
        return [elt.attr_name for elt in self.loops if elt.usage == "R"]

    def jsonschema(self) -> JSONSchema:
        """
        Emits JSON Schema for a message.

        ..  todo:: properly flatten the message to emit $loop, $segment, $composite, and $element definitions.
        """
        schema: dict[str, Any] = {
            "title": self.name,
            "description": f"xid={self.xid} name={self.name}",
        }
        if self.loops:
            schema["type"] = "object"
            schema["properties"] = {
                elt.attr_name: {"$ref": f"#/$loops/{elt.class_name}"}  # elt.jsonschema()
                for elt in self.loops
                # if elt.usage != "N"
            }
        if req := self.required:
            schema["required"] = req
        return schema


def message_reader(source: TextIO) -> Message:
    """
    See https://github.com/azoner/pyx12/blob/master/pyx12/map/map.xsd

    Note that the common (control) messages use a lot of named elements.
    The other messages use a lot of attributes.
    Hence the duplication.

    The summary of the schema is as follows:

    ::

        transaction (xid)
            name
            loop

        loop (xid, pos, usage, repeat, type)
            name
            usage
            pos
            repeat
            segment | loop

        segment (xid, usage, pos, max_use)
            name
            end_tag
            usage
            pos
            max_use
            syntax
            element | composite

        composite (xid, data_ele, usage, seq)
            data_ele
            name
            usage
            seq
            refdes
            syntax
            repeat
            element

        element (xid, data_ele, name, usage, seq)
            data_ele
            name
            usage
            seq
            refdes
            repeat
            valid_codes (external)
                code

    usage: one of R (required), S (situational), N (not used)

    syntax: uses a string with the following structure ``(E|L|P|C|R)([0-9]{2}){2,20}``

        E - Exclusion - Not more than one specified element may be present

        L - List - If first is present, at least one other must be present

        P - Paired or Multiple - If any element specified is present, all must be present

        C - Conditional - If first is present, all others must be present

        R - Required - At least one must be present

        Followed by two-digit numbers to list elements to which this syntax applies.
    """
    def element_walker(loop_context: str, base_tag: XML.Element) -> Element:
        xid_a = base_tag.attrib.get("xid")
        data_ele_a = base_tag.attrib.get("data_ele")
        usage_a = base_tag.attrib.get("usage")
        seq_a = base_tag.attrib.get("seq")
        data_ele = optional_text(base_tag, "data_ele")
        name = optional_text(base_tag, "name")
        usage = optional_text(base_tag, "usage")  # See usageType
        seq = optional_text(base_tag, "seq")
        refdes = optional_text(base_tag, "refdes")
        repeat = optional_text(base_tag, "repeat")
        if (valid_codes_tag := base_tag.find("valid_codes")) is not None:
            codes = [c.text for c in valid_codes_tag.findall("code")]
            external = valid_codes_tag.attrib.get("external")
        else:
            codes = []
            external = ""
        element_id = data_ele_a or data_ele
        # data_element = data_elements.get(element_id, None)
        # code_set = codesets.get(element_id, None)
        element = Element(
            xid=cast(str, xid_a),
            data_ele=element_id,
            name=name,
            usage=usage_a or usage,
            seq=seq_a or seq,
            refdes=refdes,
            repeat=repeat,
            codes=cast(list[str], codes),
            external=cast(str, external),
            common_def=element_id,
            loop=loop_context
        )
        logger.debug(element)
        return element

    def composite_walker(loop_context: str, base_tag: XML.Element) -> Composite:
        xid_a = base_tag.attrib.get("xid")
        data_ele_a = base_tag.attrib.get("data_ele")
        usage_a = base_tag.attrib.get("usage")
        seq_a = base_tag.attrib.get("seq")
        data_ele = optional_text(base_tag, "data_ele")
        name = optional_text(base_tag, "name")
        usage = optional_text(base_tag, "usage")  # See usageType
        seq = optional_text(base_tag, "seq")
        refdes = optional_text(base_tag, "refdes")
        syntax = [t.text for t in base_tag.findall("syntax")]  # See syntaxType rules
        repeat = optional_text(base_tag, "repeat")
        elements = [
            element_walker(loop_context, element)
            for element in base_tag.findall("element")
        ]
        composite = Composite(
            xid=cast(str, xid_a),
            data_ele=data_ele_a or data_ele,
            name=name,
            usage=usage_a or usage,
            seq=seq_a or seq,
            refdes=refdes,
            syntax=cast(list[str], syntax),
            repeat=repeat,
            elements=elements,
            loop=loop_context
        )
        logger.debug(composite)
        return composite

    def segment_walker(loop_context: str, base_tag: XML.Element):
        xid_a = base_tag.attrib.get("xid")
        pos_a = base_tag.attrib.get("pos")
        usage_a = base_tag.attrib.get("usage")
        max_use_a = base_tag.attrib.get("max_use")
        name = optional_text(base_tag, "name")
        end_tag = optional_text(base_tag, "end_tag")
        usage = optional_text(base_tag, "usage")
        pos = optional_text(base_tag, "pos")
        max_use = optional_text(base_tag, "max_use")
        syntax = [t.text for t in base_tag.findall("syntax")]  # See syntaxType rules
        # Any number of element or composite may follow...
        children: list[Element | Composite] = []
        for child_tag in base_tag.findall("*"):
            if child_tag.tag == "element":
                ec = element_walker(loop_context, child_tag)
                children.append(ec)
            elif child_tag.tag == "composite":
                cc = composite_walker(loop_context, child_tag)
                children.append(cc)
            elif child_tag.tag in {"name", "end_tag", "usage", "pos", "max_use", "syntax"}:
                pass  # Skip already processed tags
            else:
                raise RuntimeError(f"Unexpected {child_tag.tag}")
        segment = Segment(
            xid=cast(str, xid_a),
            name=name,
            end_tag=end_tag,
            usage=usage_a or usage,
            pos=pos_a or pos,
            max_use=max_use_a or max_use,
            syntax=cast(list[str], syntax),
            children=children,
            loop=loop_context
        )
        logger.debug(segment)
        return segment

    def loop_walker(base_tag: XML.Element) -> Loop:
        xid_a = cast(str, base_tag.attrib.get("xid"))
        pos_a = base_tag.attrib.get("pos")
        usage_a = base_tag.attrib.get("usage")
        repeat_a = base_tag.attrib.get("repeat")
        type_a = base_tag.attrib.get("type")  # Actually used, doesn't seem appropriate at this level.
        name = optional_text(base_tag, "name")
        usage = optional_text(base_tag, "usage")
        pos = optional_text(base_tag, "pos")
        repeat = optional_text(base_tag, "repeat")
        # Any number of loop OR segment may follow...
        children: list[Loop | Segment] = []
        for child_tag in base_tag.findall("*"):
            if child_tag.tag == "loop":
                cl = loop_walker(child_tag)
                children.append(cl)
            elif child_tag.tag == "segment":
                cs = segment_walker(pythonify(xid_a, "L"), child_tag)
                children.append(cs)
            elif child_tag.tag in {"name", "usage", "pos", "repeat"}:
                pass  # Skip already processed tags
            else:
                raise RuntimeError(f"Unexpected {child_tag.tag}")
        # TODO: DRY issue. The segment_walker (above) is given the class_name value from the Loop class (below).
        loop = Loop(
            xid=xid_a,
            name=name,
            usage=usage_a or usage,
            pos=pos_a or pos,
            repeat=repeat_a or repeat,
            type=cast(str, type_a),
            children=children,
        )
        logger.debug(loop)
        return loop

    document = XML.parse(source)
    transaction_tag = document.getroot()
    xid = transaction_tag.attrib.get("xid")
    name = cast(XML.Element, transaction_tag.find("name")).text
    loops = [
        loop_walker(loop_tag)
        for loop_tag in transaction_tag.findall("loop")
    ]
    message = Message(
        xid=cast(str, xid),
        name=cast(str, name),
        loops=loops
    )
    return message


Component: TypeAlias = Union[Element, Composite, Segment, Loop, Message]


class MessageVisitor:
    """
    Abstract Visitor for all layers of a message, in Post-Order traversal.

    This works UP the message structure from elements to composites, segments, loops,
    and finally the containing message.
    """
    def __init__(self) -> None:
        self.source_name = ""
        self.path: Deque[tuple[str, str]] = deque()
        self.codes: dict[str, Codeset]
        self.types: dict[str, DataElement]
        self.type_names: set[str]

    def codesets(self, c: dict[str, Codeset]) -> Self:
        """Setter for codesets."""
        self.codes = c
        return self

    def data_elements(self, d: dict[str, DataElement]) -> Self:
        """Setter for data elements."""
        self.types = d
        return self

    def dump_codesets(self) -> set[str]:
        return set()

    def dump_data_elements(self) -> set[str]:
        return set()

    def visit(self, component: Component) -> None:
        match component:
            case Element():
                self.element(component)
            case Composite():
                self.path.append(("COMPOSITE", component.class_name))
                for element in component.elements:
                    self.visit(element)
                self.path.pop()
                self.composite(component)
            case Segment():
                self.path.append(("SEGMENT", component.class_name))
                for seg_child in component.children:
                    self.visit(seg_child)
                self.path.pop()
                self.segment(component)
            case Loop():
                self.path.append(("LOOP", component.class_name))
                for loop_child in component.children:
                    self.visit(loop_child)
                self.path.pop()
                self.loop(component)
            case Message():
                self.path = deque([("MESSAGE", component.class_name)])
                for loop in component.loops:
                    self.visit(loop)
                self.path.pop()
                self.message(component)
            case _:
                raise RuntimeError(f"unexpected {component!r}")

    def element(self, element: Element) -> None:
        pass

    def composite(self, composite: Composite) -> None:
        pass

    def segment(self, segment: Segment) -> None:
        pass

    def loop(self, loop: Loop) -> None:
        pass

    def message(self, message: Message) -> None:
        pass


class EmitPython_Interim(MessageVisitor):
    """
    Message visitor to emit Pythonic definitions for
    Elements, Composites, Segments, Loops, and Messages.

    This is "Interim", which uses an explicit ``Schema``
    internal class definition.

    The classes written will be extension subclasses to definitions in ``x12.base``.
    """
    def dump_codesets(self) -> set[str]:
        names: set[str] = set()
        for id in self.codes:
            name = pythonify(id, 'C_')
            print(f"{name} = {self.codes[id].jsonschema()}")
            names.add(name)
        return names

    def dump_data_elements(self) -> set[str]:
        self.type_names = set()
        for name in self.types:
            name = pythonify(name, 'D_')
            print(f"{name} = {self.types[name].jsonschema()}")
            self.type_names.add(name)
        return self.type_names

    def element(self, element: Element) -> None:
        print("\n")
        print(f'class {element.class_name}(Element):\n    """{element.name}"""')
        print(f'    class Schema:')
        print(f"        json = {indent(pformat(element.jsonschema(), compact=True, sort_dicts=False), 8*' ').lstrip()}")
        if element.common_def:
            print(f"        datatype = common.{pythonify(element.common_def, 'D_')}")
        if element.external:
            print(f"        codes = common.{pythonify(element.external, 'D_')}")
        if element.codes:
            print(f"        codes = {element.codes}")
        # Size -- mostly for ISA segment, which is fixed length, ignoring the separators.
        if element.data_ele in self.types:
            if self.types[element.data_ele].min_len:
                print(f"        min_len = {self.types[element.common_def].int_min_len}")
            if self.types[element.data_ele].max_len:
                print(f"        max_len = {self.types[element.common_def].int_max_len}")
        else:
            logger.warning("no data_element for %s", element)


    def composite(self, composite: Composite) -> None:
        print("\n")
        # print(f"# {self.path}")
        print(f'class {composite.class_name}(Composite):')
        print(f'    class Schema:')
        print(f"        json = {indent(pformat(composite.jsonschema(), compact=True, sort_dicts=False), 8*' ').lstrip()}")
        for elt in composite.elements:
            # if elt.usage == "N":
            #    continue
            if elt.repeat:
                type_text = f"list[{elt.class_name}]"
            else:
                type_text = elt.class_name
            if elt.attr_name not in composite.required:
                type_text = f"{type_text} | None"
            print(f"    {elt.attr_name}: {type_text}")

    def segment(self, segment: Segment) -> None:
        print("\n")
        # print(f"# {self.path}")
        print(f'class {segment.class_name}(Segment):\n    """{segment.name}"""')
        print(f'    class Schema:')
        print(f"        json = {indent(pformat(segment.jsonschema(), compact=True, sort_dicts=False), 8*' ').lstrip()}")
        print(f"        segment_name = {segment.xid!r}")
        for elt in segment.children:
            if elt.usage == "N":
                if isinstance(elt,Composite) and len(elt.elements) == 0:
                    continue
            if elt.repeat:
                type_text = f"list[{elt.class_name}]"
            else:
                type_text = elt.class_name
            if elt.attr_name not in segment.required:
                type_text = f"{type_text} | None"
            print(f"    {elt.attr_name}: {type_text}")

    def loop(self, loop: Loop) -> None:
        print("\n")
        # print(f"# {self.path}")
        print(f'class {loop.class_name}(Loop):')
        print(f'    class Schema:')
        print(f"        json = {indent(pformat(loop.jsonschema(), compact=True, sort_dicts=False), 8*' ').lstrip()}")
        for elt in loop.children:
            # if elt.usage == "N":
            #     continue
            if elt.repeat:
                type_text = f"list[{elt.class_name}]"
            else:
                type_text = elt.class_name
            if elt.attr_name not in loop.required:
                type_text = f"{type_text} | None"
            print(f"    {elt.attr_name}: {type_text}")

    def message(self, message: Message) -> None:
        print("\n")
        print(f'class {message.class_name}(Message):\n    """{message.name}"""')
        print(f'    class Schema:')
        print(f"        json = {indent(pformat(message.jsonschema(), compact=True, sort_dicts=False), 8*' ').lstrip()}")
        for elt in message.loops:
            # if elt.usage == "N":
            #    continue
            if elt.repeat:
                type_text = f"list[{elt.class_name}]"
            else:
                type_text = elt.class_name
            if elt.attr_name not in message.required:
                type_text = f"{type_text} | None"
            print(f"    {elt.attr_name}: {type_text}")


class EmitPython_Annotated(MessageVisitor):
    """
    Message visitor to emit Pythonic definitions for
    Elements, Composites, Segments, Loops, and Messages.

    This is "Annotated". Elements don't have a separate class,
    their annotations are part of the containing Segment or Composite.

    The classes created will be extension subclasses to definitions in ``x12.base``.
    """
    def __init__(self) -> None:
        super().__init__()
        self.codes: dict[str, Codeset]
        self.types: dict[str, DataElement]

    def codesets(self, c: dict[str, Codeset]) -> Self:
        self.codes = c
        return self

    def data_elements(self, d: dict[str, DataElement]) -> Self:
        self.types = d
        return self

    def dump_codesets(self) -> set[str]:
        names: set[str] = set()
        for id, codeset in self.codes.items():
            # annotation = f"Annotated[str, Enumerated(*{codeset.values})]"
            name = pythonify(id, 'C_')
            print(f"{name} = {codeset.values!r}")
            names.add(name)
        return names

    def dump_data_elements(self) -> set[str]:
        self.type_names = set()
        for name, base_def in X12_TYPE_ANNOTATION.items():
            print(f"{name}: TypeAlias = {base_def}")
        for name, de_def in self.types.items():
            assert de_def.data_type in X12_TYPE_ANNOTATION, f"Data type {de_def.data_type!r} unknown"
            base = [str(de_def.data_type)]
            if de_def.min_len:
                base.append(f"MinLen({de_def.min_len})")
            if de_def.max_len:
                base.append(f"MaxLen({de_def.max_len})")
            name = pythonify(name, 'D_')
            print(
                f"{name}: TypeAlias = Annotated[{', '.join(base)}]"
            )
            self.type_names.add(name)
        return self.type_names

    def element(self, element: Element) -> None:
        pass  # Not a separate class in the Python.

    def composite(self, composite: Composite) -> None:
        print("\n")
        # print(f"# {self.path}")
        print(f'class {composite.class_name}(Composite):\n    """{composite.name}"""')
        if composite.elements:
            for elt in composite.elements:
                # if elt.usage == "N":
                #    continue
                alias, annotation = elt.annotation()
                # Not sure we can make optional attributes work via Union[]
                # if elt.attr_name not in composite.required:
                #     name, _, annotation = annotation.partition(":")
                #     annotation = f"{name}: Union[{annotation}, None]"
                print(f"    {alias}")
                print(f"    {annotation}")
        else:
            print("    pass")

    def segment(self, segment: Segment) -> None:
        print("\n")
        # print(f"# {self.path}")
        print(f'class {segment.class_name}(Segment):\n    """{segment.name}"""\n    _segment_name = {segment.attr_name.upper()!r}')
        if segment.children:
            for elt in segment.children:
                if elt.usage == "N":
                    if isinstance(elt,Composite) and len(elt.elements) == 0:
                        continue
                alias, annotation = elt.annotation()
                # Not sure we can make optional attributes work via Union[]
                # if elt.attr_name not in segment.required:
                #     name, _, annotation = annotation.partition(":")
                #     annotation = f"{name}: Union[{annotation}, None]"
                print(f"    {alias}")
                print(f"    {annotation}")
        else:
            print("    pass")

    def loop(self, loop: Loop) -> None:
        print("\n")
        # print(f"# {self.path}")
        print(f'class {loop.class_name}(Loop):')
        if loop.children:
            for elt in loop.children:
                # if elt.usage == "N":
                #    continue
                alias, annotation = elt.annotation()
                # Not sure we can make optional attributes work via Union[]
                # if elt.attr_name not in loop.required:
                #     name, _, annotation = annotation.partition(":")
                #     annotation = f"{name}: Union[{annotation}, None]"
                print(f"    {alias}")
                print(f"    {annotation}")
        else:
            print("    pass")

    def message(self, message: Message) -> None:
        print("\n")
        print(f'class {message.class_name}(Message):\n    """{message.name}"""')
        for elt in message.loops:
            # if elt.usage == "N":
            #    continue
            alias, annotation = elt.annotation()
            # Not sure we can make optional attributes work via Union[]
            # if elt.attr_name not in message.required:
            #     name, _, annotation = annotation.partition(":")
            #     annotation = f"{name}: Union[{annotation}, None]"
            print(f"    {alias}")
            print(f"    {annotation}")

class PythonMaker:
    visitor_class: Type[MessageVisitor]

    def make_common_module(
            self,
            codesets: dict[str, Codeset],
            data_elements: dict[str, DataElement],
            target_package: Path
    ) -> set[str]:
        """
        Emits common module with definitions of common data types.

        Depends on ``EmitPython_Interim`` class.
        """
        target_path = (target_package / "common").with_suffix(".py")

        cls = self.visitor_class
        code_writer = cls().data_elements(data_elements).codesets(codesets)
        names: set[str] = set()

        with target_path.open('w') as target_file:
            with redirect_stdout(target_file):
                print('"""')
                print(f"Created {datetime.datetime.now()}")
                print('"""')
                print()
                print("from typing import TypeAlias, Annotated")
                print("import datetime")
                print("from decimal import Decimal")
                print("from x12.annotations import *")
                names.union(code_writer.dump_codesets())
                names.union(code_writer.dump_data_elements())
        return names

    def make_message_module(
            self,
            codesets: dict[str, Codeset],
            data_elements: dict[str, DataElement],
            source_path: Path,
            target_package: Path
    ) -> None:
        """
        Emits a module to parse a message.

        Depends on ``EmitPython_Interim`` class.
        """
        target_stem = f'msg_{source_path.stem.replace(".", "_")}'
        target_path = (target_package / target_stem).with_suffix(".py")
        logger.info("%s -> %s", source_path.name, target_path.name)

        with source_path.open() as source:
            msg = message_reader(source)

        cls = self.visitor_class
        code_writer = cls().data_elements(data_elements).codesets(codesets)
        code_writer.source_name = source_path.stem

        with target_path.open('w') as target_file:
            with redirect_stdout(target_file):
                print('"""')
                print(source_path.stem)
                print(f"Created {datetime.datetime.now()}")
                # print(msg)
                print('"""')
                print("from .base import *")
                print("from .common import *")
                code_writer.visit(msg)


    def make_python(self, base: Path, target: Path) -> None:
        """
        Extract from XML schema, creating Python.

        Each message becomes a separate module in an overall "x12" package.

        Depends on EmitPython_Interim
        """
        with (base / "codes.xml").open() as source:
            logger.info("Reading %s", source.name)
            codesets = {
                codeset.id: codeset
                for codeset in code_reader(source)
            }
        with (base / "dataele.xml").open() as source:
            logger.info("Reading %s", source.name)
            data_elements = {
                data_element.ele_num: data_element
                for data_element in element_reader(source)
            }
        names = self.make_common_module(codesets, data_elements, x12_package)

        for source_path in sorted(base.glob("x12.control.*.xml")):
            self.make_message_module(codesets, data_elements, source_path, x12_package)

        for source_path in sorted(base.glob("[0-9][0-9][0-9]*.xml")):
            self.make_message_module(codesets, data_elements, source_path, x12_package)

        with (base / "maps.xml").open() as source:
            maps = list(map_reader(source))

        # TODO: Emit the mapping objects.
        # self.make_mapper(maps, message_modules)


class InterimPythonMaker(PythonMaker):
    """Make Python using Schema constructs."""
    visitor_class = EmitPython_Interim


class AnnotatedPythonMaker(PythonMaker):
    """Make Python using Annotated definitions."""
    visitor_class = EmitPython_Annotated


## TODO: Combine make_message_schema() and  make_jsonschema() into a class

def make_message_schema(
        codesets: dict[str, Codeset],
        data_elements: dict[str, DataElement],
        source_path: Path,
        target_package: Path
) -> None:
    """
    Emit JSON Schema for a message.

    ..  todo:: Include the "$ref" data types and code sets.

    JSON Schema **should** be created from the Python, not from the XML.
    """
    target_stem = f'msg_{source_path.stem.replace(".", "_")}'
    target_path = (target_package / target_stem).with_suffix(".json")
    logger.info("%s -> %s", source_path.name, target_path.name)
    with source_path.open() as source:
        msg = message_reader(source)
    print(msg.jsonschema())


def make_schema(base: Path, target: Path) -> None:
    """
    Extract from XML schema, creating JSON Schema.

    Each message becomes a separate .json document.

    JSON Schema **should** be created from the Python, not from the XML.
    """
    with (base / "codes.xml").open() as source:
        logger.info("Reading %s", source.name)
        codesets = {
            codeset.id: codeset
            for codeset in code_reader(source)
        }
    with (base / "dataele.xml").open() as source:
        logger.info("Reading %s", source.name)
        data_elements = {
            data_element.ele_num: data_element
            for data_element in element_reader(source)
        }
    for source_path in sorted(base.glob("x12.control.*.xml")):
        make_message_schema(codesets, data_elements, source_path, x12_package)

    for source_path in sorted(base.glob("[0-9][0-9][0-9]*.xml")):
        make_message_schema(codesets, data_elements, source_path, x12_package)


def make_python(maker: PythonMaker, base: Path, x12_package: Path) -> None:
    """
    Extract from XML schema and make Python modules for parsing messages.
    """
    logger.info("Source %s", base)
    logger.info("Creating %s", x12_package)
    # maker = InterimPythonMaker()
    # maker = AnnotatedPythonMaker()
    maker.make_python(base, x12_package)

def make_jsonschema(base: Path, json_dir: Path) -> None:
    """
    Making JSON Schema is possible.
    JSON Schema **should** be created from the Python, not directly from the XML.
    """
    logger.info("Source %s", base)
    logger.info("Creating %s", json_dir)
    make_schema(base, json_dir)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    base = Path.home() / "github" / "pyx12" / "pyx12" / "map"
    output = "python"

    if output == "python":
        x12_package = Path.cwd().parent / "x12"
        make_python(AnnotatedPythonMaker(), base, x12_package)
    elif output == "schema-python":
        x12_package = Path.cwd().parent / "x12"
        make_python(InterimPythonMaker(), base, x12_package)
    elif output == "json":
        json_dir = Path.cwd() / "json"
        make_jsonschema(base, json_dir)
    else:
        raise ValueError("unknown output option, {output}")
    logging.shutdown()



"""
x12.control.00401
Created 2023-05-19 10:17:35.812747
"""
from .base import *
from .common import *


class ISA_LOOP_ISA(Segment):
    """Interchange Control Header"""
    _segment_name = 'ISA'
    
    isa01: Annotated[I01, Title('Authorization Information Qualifier'), Usage('R'), Position(1), Enumerated(*['00', '03'])]
    
    isa02: Annotated[I02, Title('Authorization Information'), Usage('R'), Position(2)]
    
    isa03: Annotated[I03, Title('Security Information Qualifier'), Usage('R'), Position(3), Enumerated(*['00', '01'])]
    
    isa04: Annotated[I04, Title('Security Information'), Usage('R'), Position(4)]
    
    isa05: Annotated[I05, Title('Interchange ID Qualifier'), Usage('R'), Position(5), Enumerated(*['01', '14', '20', '27', '28', '29', '30', '33', 'ZZ'])]
    
    isa06: Annotated[I06, Title('Interchange Sender ID'), Usage('R'), Position(6)]
    
    isa07: Annotated[I05, Title('Interchange ID Qualifier'), Usage('R'), Position(7), Enumerated(*['01', '14', '20', '27', '28', '29', '30', '33', 'ZZ'])]
    
    isa08: Annotated[I07, Title('Interchange Receiver ID'), Usage('R'), Position(8)]
    
    isa09: Annotated[I08, Title('Interchange Date'), Usage('R'), Position(9)]
    
    isa10: Annotated[I09, Title('Interchange Time'), Usage('R'), Position(10)]
    
    isa11: Annotated[I10, Title('Interchange Control Standards Identifier'), Usage('R'), Position(11), Enumerated(*['U'])]
    
    isa12: Annotated[I11, Title('Interchange Control Version Number'), Usage('R'), Position(12), Enumerated(*['00400', '00401', '00402'])]
    
    isa13: Annotated[I12, Title('Interchange Control Number'), Usage('R'), Position(13)]
    
    isa14: Annotated[I13, Title('Acknowledgment Requested'), Usage('R'), Position(14), Enumerated(*['0', '1'])]
    
    isa15: Annotated[I14, Title('Usage Indicator'), Usage('R'), Position(15), Enumerated(*['P', 'T'])]
    
    isa16: Annotated[I15, Title('Component Element Separator'), Usage('R'), Position(16)]


class GS_LOOP_GS(Segment):
    """Functional Group Header"""
    _segment_name = 'GS'
    
    gs01: Annotated[D_479, Title('Functional Identifier Code'), Usage('R'), Position(1), Enumerated(*['HN', 'HR', 'HS', 'HB', 'HI', 'RA', 'BE', 'HP', 'HC', 'FA'])]
    
    gs02: Annotated[D_142, Title("Application Sender's Code"), Usage('R'), Position(2)]
    
    gs03: Annotated[D_124, Title("Application Receiver's Code"), Usage('R'), Position(3)]
    
    gs04: Annotated[D_373, Title('Date'), Usage('R'), Position(4)]
    
    gs05: Annotated[D_337, Title('Time'), Usage('R'), Position(5)]
    
    gs06: Annotated[D_28, Title('Group Control Number'), Usage('R'), Position(6)]
    
    gs07: Annotated[D_455, Title('Responsible Agency Code'), Usage('R'), Position(7), Enumerated(*['X'])]
    
    gs08: Annotated[D_480, Title('Version / Release / Industry Identifier Code'), Usage('R'), Position(8), Enumerated(*['004010', '004010X070', '004010X093', '004010X092', '004010X094', '004010X061', '004010X095', '004010X091', '004010X096', '004010X097', '004010X098', '004010X093A1', '004010X092A1', '004010X094A1', '004010X061A1', '004010X095A1', '004010X091A1', '004010X096A1', '004010X097A1', '004010X098A1'])]


class GS_LOOP_GE(Segment):
    """Functional Group Trailer"""
    _segment_name = 'GE'
    
    ge01: Annotated[D_97, Title('Number of Transaction Sets Included'), Usage('R'), Position(1)]
    
    ge02: Annotated[D_28, Title('Group Control Number'), Usage('R'), Position(2)]


class GS_LOOP(Loop):
    
    gs: Annotated[GS_LOOP_GS, Title('Functional Group Header'), Usage('R'), Position(10), Required(True)]
    
    ge: Annotated[GS_LOOP_GE, Title('Functional Group Trailer'), Usage('R'), Position(30), Required(True)]


class ISA_LOOP_TA1(Segment):
    """Interchange Acknowledgement"""
    _segment_name = 'TA1'
    
    ta101: Annotated[I12, Title('Interchange Control Number'), Usage('R'), Position(1)]
    
    ta102: Annotated[I08, Title('Interchange Date'), Usage('R'), Position(2)]
    
    ta103: Annotated[I09, Title('Interchange Time'), Usage('R'), Position(3)]
    
    ta104: Annotated[I17, Title('Interchange Acknowledgement Code'), Usage('R'), Position(4), Enumerated(*['A', 'E', 'R'])]
    
    ta105: Annotated[I18, Title('Interchange Note Code'), Usage('R'), Position(5), Enumerated(*['000', '001', '002', '003', '004', '005', '006', '007', '008', '009', '010', '011', '012', '013', '014', '015', '016', '017', '018', '019', '020', '021', '022', '023', '024', '025', '026', '027', '028', '029', '030', '031'])]


class ISA_LOOP_IEA(Segment):
    """Interchange Control Trailer"""
    _segment_name = 'IEA'
    
    iea01: Annotated[I16, Title('Number of Included Functional Groups'), Usage('R'), Position(1)]
    
    iea02: Annotated[I12, Title('Interchange Control Number'), Usage('R'), Position(2)]


class ISA_LOOP(Loop):
    
    isa: Annotated[ISA_LOOP_ISA, Title('Interchange Control Header'), Usage('R'), Position(10), Required(True)]
    ItemGs_Loop: TypeAlias = Annotated[GS_LOOP, Title('Functional Group Header'), Usage('R'), Position(20), Required(True)]
    gs_loop: Annotated[list[ItemGs_Loop], MinItems(1)]
    
    ta1: Annotated[ISA_LOOP_TA1, Title('Interchange Acknowledgement'), Usage('S'), Position(20), Required(True)]
    
    iea: Annotated[ISA_LOOP_IEA, Title('Interchange Control Trailer'), Usage('R'), Position(30), Required(True)]


class X12(Message):
    """X12 Control"""
    ItemIsa_Loop: TypeAlias = Annotated[ISA_LOOP, Title('Interchange Control Header'), Usage('R'), Position(1), Required(True)]
    isa_loop: Annotated[list[ItemIsa_Loop], MinItems(1)]



"""
x12
====

The bulk of this package is defined by extraction from other sources.
Specifically the :py:mod:`tools.xml_extract` module.

The message definition all rely on the :py:mod:`x12.base`
and :py:mod:`x12.annotations` modules.

The tools build an :py:mod:`x12.common` module.
Each message is parsed by a separate module.
"""

from .base import Source
